#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <semaphore.h>

#include <usbpromethean.h>

#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>

sem_t *sem = NULL;
#include <signal.h>
void termination_handler (int signum)
{
sem_t *m = sem;
    printf("Interrupted\n");
    sem = NULL;
    sem_close(m);
    _exit(-1);
}

// load calibration on board (not in memory)
int LoadCalibration(const char *szDevice)
{
    int board_descriptor = -1;
    int f = -1;
    char szFile[256] = {"Unknown"};
    unsigned char data[20];
    int retval = 0, nread = 0;
    const int TimeoutMS = 5000;

    printf("Promethean: Opening calibration for device %s\n",szDevice);
    board_descriptor = open(szDevice,O_RDWR);
    if (board_descriptor < 0)
    {
        perror("Error opening ACTIVBoard: ");
        return -1;
    }

    if (!ioctl(board_descriptor,ACTIV_SET_TIMEOUT,&TimeoutMS))
    {
        // send :Q1 to identify the device
        unsigned char identified = 0, type = 0;
        int retries = 0;

        while ((!identified) && (++retries < 10))
        {
            ioctl(board_descriptor,ACTIV_FLUSH_QUEUE,NULL);
            write(board_descriptor,":Q1\r",4);
            memset(data,0,sizeof(data));
            identified = 1;


            if (read(board_descriptor,data,20) >= 0)
            {
                type = (unsigned char)atoi(data+5);
                printf("Promethean: Device type %x\n",type);
                switch (type)
                {
                case 3:
                case 4:
                case 7:
                case 8:

                    memset(data,0,sizeof(data));
                    identified = 0;
                    write(board_descriptor,":SN\r",4);
                    nread = read(board_descriptor,data,20);
                    printf("Promethean: Received %d bytes from :SN\n",nread);
                    printf("Promethean: 0x%02x 0x%02x\n",data[0],data[1]);

                    if (nread >= 0)
                    {
                        if (data[1] == 0x69)
                        {
                            identified = 1;
                            switch (type)
                            {
                            case 8:
                                if (data[2] + data[3] + data[4] + data[5])
                                {
                                    sprintf(szFile,"/etc/promethean/calibration/calibrate-%02x%02x%02x%02x%02x%02x.data",
                                            data[7],data[6],
                                            data[2],data[3],data[4],data[5]);
                                }
                                else
                                {
                                    identified = 0;
                                }
                                break;
                            default:
                                sprintf(szFile,"/etc/promethean/calibration/calibrate-%02x%02x%02x%02x%02x%02x.data",
                                        data[7],data[6],
                                        data[5],data[4],data[3],data[2]);
                                break;
                            }
                        }
                    }
                    break;
                case 5:
                    memset(data,0,sizeof(data));
                    write(board_descriptor,":HI\r",4);
                    nread = read(board_descriptor,data,20);

                    printf("Promethean: Received %d bytes from :HI\n",nread);
                    printf("Promethean: 0x%02x 0x%02x\n",data[0],data[1]);
                    identified = 0;

                    if (nread >= 0)
                    {
                        if (data[1] == 0x77)
                        {
                            identified = 1;
                            sprintf(szFile,"/etc/promethean/calibration/calibrate-000000%02x%02x%02x.data",
                                    data[2],data[3],data[4]);
                        }
                    }
                    break;
		case 0xe: 
			retval = -3;
			retries = 10;
			identified=0;
		break;
                default:
                    fprintf(stderr,"Promethean: (activlc) Invalid device type.");
                    retval = -2;
                    identified = 0;
                    break;

                }
            }
            if (!identified)
                usleep(500000 * ((retries % 2) +1));
        }

        if (identified)
        {
            // open the calibration file
            f = open(szFile,O_RDONLY);

            if (f < 0)
            {
                close(board_descriptor);
                perror("Promethean: Error opening calibration file: ");
                fprintf(stderr,"%s\n",szFile);
                return -3;
            }

            CALIB_DATA d;
            if (read(f,&d,sizeof(d)) == sizeof(d))
            {
                int io = ioctl(board_descriptor,ACTIV_SET_CALIBRATION_DATA,&d);
                if (io)
                    perror("Promethean: Error sending calibration data: ");
                retval = io?-4:0;
            }
            else
            {
                fprintf(stderr,"Promethean: Invalid calibration data size.");
                retval = -5;
            }
        }
        else
        {
            fprintf(stderr,"Promethean: Cannot identify device.");
            retval = -6;
        }

        close(f);
    }
    else
    {
        fprintf(stderr,"Promethean: Cannot set timeout.");
        retval = -7;
    }

    close(board_descriptor);
    return retval;
}

int processArgs( int argc, char **argv )
{
    int ch;
    /* options descriptor */
    static struct option longopts[] = {
        { "load",      1,      NULL,           'l' },
        { NULL,         0,                      NULL,           0 }
    };


    // when plugging boards, all devices get loaded with --load ... --device ...
    // this takes precedence over the normal device detection.

    // cmd line args return straight away
    while ((ch = getopt_long(argc, argv, "l:", longopts, NULL)) != -1)
    {
        switch (ch)
        {
        case 'l':
            if (optarg)
                return LoadCalibration(optarg);
            break;
        }
    }

    return -1;
}

int lock()
{
    int ret = -1;
    sem = sem_open("activlc",O_CREAT,S_IRUSR|S_IWUSR,1);

    if (sem)
    {
        if (sem_trywait(sem) < 0)
        {
            usleep(15000 * 1000);
            if (sem)

                if (sem_trywait(sem) >= 0)
                    ret = 0;
        }
        else
            ret = 0;
    }

    if (ret < 0)
        sem_unlink("activlc");

    return ret;
}

void unlock()
{
    if (sem)
    {
        sem_post(sem);
        sem_close(sem);
    }
}

int main( int argc, char **argv )
{
    int ret = -1;

    struct sigaction new_action, old_action;

    /* Set up the structure to specify the new action. */
    new_action.sa_handler = termination_handler;
    sigemptyset (&new_action.sa_mask);
    new_action.sa_flags = 0;

    sigaction (SIGINT, NULL, &old_action);
    if (old_action.sa_handler != SIG_IGN)
        sigaction (SIGINT, &new_action, NULL);
    sigaction (SIGHUP, NULL, &old_action);
    if (old_action.sa_handler != SIG_IGN)
        sigaction (SIGHUP, &new_action, NULL);
    sigaction (SIGTERM, NULL, &old_action);
    if (old_action.sa_handler != SIG_IGN)
        sigaction (SIGTERM, &new_action, NULL);

    if (lock() == 0)
    {
        ret = processArgs(argc,argv);
        unlock();
    }

    return ret;
}
