/*
* xf86Promethean
*
* This driver assumes Linux Input Device support, available for USB devices.
*
* Copyright 2006 by Promethean Technologies Group Ltd.
*
* Lineage: This driver is based on both the xf86HyperPen and xf86Wacom tablet
*          drivers.
*
*      xf86HyperPen -- modified from xf86Summa (c) 1996 Steven Lang
*          (c) 2000 Roland Jansen
*          (c) 2000 Christian Herzog (button & 19200 baud support)
*
*      xf86Wacom -- (c) 1995-2001 Frederic Lepied
*
*
* fPermission to use, copy, modify, distribute, and sell this software and its
* documentation for any purpose is hereby granted without fee, provided that
* the above copyright notice appear in all copies and that both that
* copyright notice and this permission notice appear in supporting
* documentation, and that the name of Bryan W. Headley not be used in
* advertising or publicity pertaining to distribution of the software
* without specific, written prior permission.  Bryan W. Headley makes no
* representations about the suitability of this software for any purpose.
* It is provided "as is" without express or implied warranty.
*
*/

#include <time.h>
#include <fcntl.h>
#include <pthread.h>
#include <memory.h>
#include "xf86promethean.h"

#define DRIVER_MAJOR 5
#define DRIVER_MINOR 4

/* we rely on stat to track changes to devices. xorg redefines stat to xf86stat, which
* doesn't have the data we need. */
#ifdef stat
#undef stat
#endif

#include <sys/stat.h>



/* These are the parameter settings, and their default values.
* We use this to make a 'fakeLocal'...
*/
static
        const char *default_options[] =
{
    "BaudRate",     "9600",
    "StopBits",     "1",
    "DataBits",     "8",
    "Parity",       "None",
    "VMin",         "1",
    "Vtime",        "10",
    "FlowControl",  "Xoff",
    NULL
};

/***********************************************************************
static int
project(double value, double min, double max, int norm)
{
return (norm - 1) * ((value - min) / (max - min)) + 0.5;
}
*/

/***********************************************************************
* PrometheanIsValidFileDescriptor
*
* Determines whether a given fd is valid/open by abusing
* the UNIX notion that anything >= 0 is valid, and our
* vapid notion that VALUE_NA is a legal value for a closed
* handle. But it does get rid of stupidity of 'if (fd < 0)', which
* defeats the purpose of a #define...
*/
static Bool
        PrometheanIsValidFileDescriptor(
                int fd)
{
    if (fd == VALUE_NA)
    {
        return BadMatch;
    }
    return Success;
}

/***********************************************************************
* PrometheanDispatcher
*
* Call dispatcher for this driver.
*/
static Bool
        PrometheanDispatcher(
                DeviceIntPtr dev,
                int          requestCode)
{
    const char*     methodName = "PrometheanDispatcher";
    LocalDevicePtr  local  = (LocalDevicePtr)dev->public.devicePrivate;
    PrometheanDevicePtr device = (PrometheanDevicePtr)PRIVATE(dev);

    CHECK_DEVICE_INT_PTR(dev, __LINE__);
    CHECK_LOCAL_DEVICE_PTR(local, __LINE__);
    CHECK_PROMETHEAN_DEVICE_PTR(device, __LINE__);

    DBG(10, xf86Msg(X_INFO, "%s: entry\n", methodName));

    switch (requestCode)
    {
    case DEVICE_INIT:
        {
            DBG(15, xf86Msg(X_INFO,
                            "%s: type=%s DeviceIntPtr.id=%d request=DEVICE_INIT\n",
                            methodName,
                            (DEVICE_ID(device->deviceId) == STYLUS_ID) ? "Stylus" :
                            (DEVICE_ID(device->deviceId) == CURSOR_ID) ? "Cursor" :
                            (DEVICE_ID(device->deviceId) == ERASER_ID) ? "Eraser" :
                            "!UNKNOWN!",
                            dev->id));

            if (PrometheanAllocateDriverClasses(dev) == BadMatch ||
                PrometheanOpenDriver(dev)            == BadMatch)
            {
                return BadMatch;
            }
        }
        break;

    case DEVICE_ON:
        {
            DBG(10, xf86Msg(X_INFO, "%s device type=%s, request=DEVICE_ON\n",
                            methodName,
                            (DEVICE_ID(device->deviceId) == STYLUS_ID) ? "Stylus" :
                            (DEVICE_ID(device->deviceId) == CURSOR_ID) ? "Cursor" :
                            (DEVICE_ID(device->deviceId) == ERASER_ID) ? "Eraser" :
                            "!UNKNOWN!"));

            if (PrometheanIsValidFileDescriptor(local->fd) == BadMatch &&
                PrometheanOpenDriver(dev)                  == BadMatch)
            {
                xf86Msg(X_INFO, "%s: type=%s, Unable to open Promethean Device\n",
                        methodName,
                        (DEVICE_ID(device->deviceId) == STYLUS_ID) ? "Stylus" :
                        (DEVICE_ID(device->deviceId) == CURSOR_ID) ? "Cursor" :
                        (DEVICE_ID(device->deviceId) == ERASER_ID) ? "Eraser" :
                        "!UNKNOWN!");
                dev->inited = FALSE;
                return BadMatch;
            }
            if (PrometheanIsValidFileDescriptor(local->fd) == Success)
            {
                xf86AddEnabledDevice(local);
                dev->public.on = TRUE;
            }
        }
        break;

    case DEVICE_OFF:
        {
            DBG(10, xf86Msg(X_INFO, "%s: DEVICE_OFF: device type=%s\n",
                            methodName,
                            (DEVICE_ID(device->deviceId) == STYLUS_ID) ? "Stylus" :
                            (DEVICE_ID(device->deviceId) == CURSOR_ID) ? "Cursor" :
                            (DEVICE_ID(device->deviceId) == ERASER_ID) ? "Eraser" :
                            "!UNKNOWN!"));

            if (PrometheanIsValidFileDescriptor(local->fd) == Success)
            {
                xf86RemoveEnabledDevice(local);
                PrometheanCloseDriver(local);
            }
            dev->public.on = FALSE;
        }
        break;

    case DEVICE_CLOSE:
        {
            DBG(10, xf86Msg(X_INFO, "%s DEVICE_CLOSE, device type=%s\n",
                            methodName,
                            (DEVICE_ID(device->deviceId) == STYLUS_ID) ? "Stylus" :
                            (DEVICE_ID(device->deviceId) == CURSOR_ID) ? "Cursor" :
                            (DEVICE_ID(device->deviceId) == ERASER_ID) ? "Eraser" :
                            "!UNKNOWN!"));

            if (PrometheanIsValidFileDescriptor(local->fd) == Success)
            {
                xf86RemoveEnabledDevice(local);
                PrometheanCloseDriver(local);
            }
            dev->public.on = FALSE;
        }
        break;

    default:
        {
            xf86Msg(X_INFO, "%s: device type=%s, Unsupported request=%d\n",
                    methodName,
                    (DEVICE_ID(device->deviceId) == STYLUS_ID) ? "Stylus" :
                    (DEVICE_ID(device->deviceId) == CURSOR_ID) ? "Cursor" :
                    (DEVICE_ID(device->deviceId) == ERASER_ID) ? "Eraser" :
                    "!UNKNOWN!",
                    requestCode);
            return BadMatch;
        }
        break;
    }
    DBG(10, xf86Msg(X_INFO, "%s: type=%s, ends successfully\n",
                    methodName,
                    (DEVICE_ID(device->deviceId) == STYLUS_ID) ? "Stylus" :
                    (DEVICE_ID(device->deviceId) == CURSOR_ID) ? "Cursor" :
                    (DEVICE_ID(device->deviceId) == ERASER_ID) ? "Eraser" :
                    "!UNKNOWN!"));
    return Success;
}

/***********************************************************************
* PrometheanAllocateDriverStructs --
*
* Helper to PrometheanDispatcher -- performs the 'DEVICE_INIT' function.
*/
static Bool
        PrometheanAllocateDriverClasses(
                DeviceIntPtr dev)
{
    const char* methodName = "PrometheanAllocateDeviceStructs";
    LocalDevicePtr  local  = (LocalDevicePtr)dev->public.devicePrivate;
    PrometheanDevicePtr device = (PrometheanDevicePtr)PRIVATE(dev);
    CARD8   map[32];
    const int     numAxes = 3;
    const int     numButtons = 3;
    int     loop;
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
    Atom *ButtonLabels = NULL, *AxesLabels = NULL;
#endif
    CHECK_DEVICE_INT_PTR(dev, __LINE__);
    CHECK_LOCAL_DEVICE_PTR(local, __LINE__);
    CHECK_PROMETHEAN_DEVICE_PTR(device, __LINE__);

    DBG(20, xf86Msg(X_INFO, "%s: begins, device=%s\n", methodName, local->name));

    /* numAxes    = 3; */            /* X, Y, Z*/
    /* numButtons = 3; */            /* left, right */

    memset(map, 0, sizeof(map));

#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
	ButtonLabels = xcalloc(3, sizeof(Atom));
    ButtonLabels[0] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_LEFT);
    ButtonLabels[1] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_RIGHT);
    ButtonLabels[2] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_MIDDLE);

	AxesLabels = xcalloc(3, sizeof(Atom));
    AxesLabels[0] = XIGetKnownProperty(AXIS_LABEL_PROP_ABS_X);
    AxesLabels[1] = XIGetKnownProperty(AXIS_LABEL_PROP_ABS_Y);
    AxesLabels[2] = XIGetKnownProperty(AXIS_LABEL_PROP_ABS_Z);
#endif
    for (loop=1; loop <= numButtons; ++loop)
    {
        map[loop] = loop;
    }

    /* These routines are typedef'd for Bool, yet return
* TRUE|FALSE, which is polar opposite of definition
* for Success|BadMatch (1,0 versus 0,1)
*/
    if (InitButtonClassDeviceStruct(dev, numButtons,
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
                                    ButtonLabels,
#endif
                                    map)
        == FALSE ||
                InitFocusClassDeviceStruct(dev)
        == FALSE ||
                InitProximityClassDeviceStruct(dev)
        == FALSE ||
                InitValuatorClassDeviceStruct(dev, numAxes,
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
                                              AxesLabels,
#endif
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 3
                                              xf86GetMotionEvents,
#endif
                                              local->history_size,
                                              ((device->flags & ABSOLUTE_FLAG) != 0 ? Absolute : Relative)
                                              | OutOfProximity)
        == FALSE )
        {
        xf86Msg(X_INFO, "%s: abends: Cannot allocate Class Device Struct(s)\n",
                methodName);
        return BadMatch;
    }

    /* Allocate the motion history buffer if needed
*/
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
    xf86MotionHistoryAllocate(local);
#endif

    DBG(20, xf86Msg(X_INFO, "%s: ends successfully\n", methodName));

    return Success;
}

/***********************************************************************
* PrometheanUninit --
*
* Called when the driver instance is unloaded.
*/
static void
        PrometheanUninit(
                InputDriverPtr    drv,
                LocalDevicePtr    local,
                int               flags)
{
    const char*     methodName = "PrometheanUninit";
    PrometheanDevicePtr device = (PrometheanDevicePtr) local->private;

    DBG(10, xf86Msg(X_INFO, "%s: entry\n", methodName));

    CHECK_INPUT_DRIVER_PTR(drv, __LINE__);
    CHECK_LOCAL_DEVICE_PTR(local, __LINE__);
    CHECK_PROMETHEAN_DEVICE_PTR(device, __LINE__);


#ifndef HAL_HOTPLUG
    DBG(10, xf86Msg(X_INFO, "%s: HAL_HOTPLUG NOT DEFINED\n", methodName));
    /*
* Don't know why, but this crashes with Xorg 7.4 using HAL_HOTPLUG.
*/
    if (device->useHalInput == FALSE)
        if (PrometheanDispatcher(local->dev, DEVICE_OFF) == BadMatch)
        {
        return;
    }
#endif

    if (device != NULL)
    {
        if (device->devicePath != NULL)
        {
            xfree(device->devicePath);
            device->devicePath = NULL;
        }
        xfree(device);

        /* do not set device to null, set the local object */
        local->private = NULL;
    }
    xf86DeleteInput(local, 0);

    DBG(10, xf86Msg(X_INFO, "%s: ends successfully\n", methodName));
}

/***********************************************************************
* xf86PrometheanInit --
*
* Called when the module subsection is found in XF86Config4.
* This parses the parameters found in the given section of the file.
*/
static InputInfoPtr
        PrometheanInit(
                InputDriverPtr    drv,
                IDevPtr           dev,
                int               flags)
{
    const char*       methodName = "PrometheanInit";
    LocalDevicePtr    local      = NULL;
    LocalDevicePtr    fakeLocal;
    char *ptr;
    PrometheanDevicePtr   device;
    char*             devicePath;
    char*             tempStr;

    CHECK_INPUT_DRIVER_PTR(drv, __LINE__);
    CHECK_I_DEV_PTR(dev, __LINE__);

    xf86Msg(X_INFO, "%s: begins for '%s'\n", methodName, dev->identifier);

    fakeLocal = (LocalDevicePtr) xcalloc(1, sizeof(LocalDeviceRec));
    if (fakeLocal == NULL)
    {
        xf86Msg(X_INFO, "%s: cannot allocate a LocalDevicePtr struct\n",
                methodName);
        return NULL;
    }

    fakeLocal->conf_idev = dev;

    /* We use fakeLocal here because we need something to parse
* what driver type is being asked for (stylus, eraser, cursor.)
* But, building a 'real' LocalDevicePtr means knowing which type
* of device we're building, which we don't yet know :-) Hence,
* the 'fake' struct.
*/
    xf86CollectInputOptions(fakeLocal, default_options, NULL);

    /* Device
*/

    ptr = fakeLocal->options;
    do
    {
        if (ptr)
        {
            xf86Msg(X_INFO, "%s: Option %s Value %s.\n",
                    methodName,
                    xf86OptionName(ptr),
                    xf86OptionValue(ptr));
        }
    } while ( (ptr = xf86NextOption(ptr)) );

    devicePath = xf86FindOptionValue(fakeLocal->options, "Device");
    /* try path... */
    if (devicePath == NULL)
    {
        devicePath = xf86FindOptionValue(fakeLocal->options, "path");
        if (devicePath)
            xf86AddNewOption(fakeLocal->options, "Device", devicePath);
    }
    if (devicePath == NULL)
    {
        xf86Msg(X_INFO, "%s: '%s': No Device specified.\n",
                methodName,
                dev->identifier);
        goto SetupProc_fail;
    }

    /* Type. Possible extension: multiple pens -> different roles. One pen stylus, one eraser ?
*/
    tempStr = xf86FindOptionValue(fakeLocal->options, "Type");
    if (tempStr != NULL && xf86NameCmp(tempStr, "stylus") == 0)
    {
        local = PrometheanAllocateLocal(drv, dev->identifier, devicePath,STYLUS_ID, ABSOLUTE_FLAG);
    }
    else if (tempStr != NULL && xf86NameCmp(tempStr, "cursor") == 0)
    {
        local = PrometheanAllocateLocal(drv, dev->identifier, devicePath,
                                        CURSOR_ID, ABSOLUTE_FLAG);
    }
    else if (tempStr != NULL && xf86NameCmp(tempStr, "eraser") == 0)
    {
        local = PrometheanAllocateLocal(drv, dev->identifier, devicePath,
                                        ERASER_ID, ABSOLUTE_FLAG);
    }
    else if (tempStr != NULL && xf86NameCmp(tempStr, "rubber") == 0)
    {
        local = PrometheanAllocateLocal(drv, dev->identifier, devicePath,
                                        ERASER_ID, ABSOLUTE_FLAG);
    }
    else
    {
        local = PrometheanAllocateLocal(drv, dev->identifier, devicePath,STYLUS_ID, ABSOLUTE_FLAG);
        xf86Msg(X_INFO, "%s: '%s': No type or invalid type specified.default to 'stylus'\n",
                methodName,	dev->identifier);
    }

    /* Capture errors allocating LocalDevicePtr and PrometheanDevicePtr
*/
    if (local == NULL)
    {
        xfree(fakeLocal);
        return NULL;
    }

    /* Paranoia is a good thing :-)
*/
    device = (PrometheanDevicePtr) local->private;
    if (device == NULL)
    {
        xfree(local);
        return NULL;
    }

    /* Move the options from fakeLocal over to local, now that we
* have a 'real' LocalDevicePtr.
*/
    local->options   = fakeLocal->options;
    local->conf_idev = fakeLocal->conf_idev;
    local->name      = dev->identifier;
    xfree(fakeLocal);

    /* If there are no other devices sharing this device, then we also have
* no PrometheanCommonPtr, either. So, let's build one of those..
*/
    device->common = PrometheanAssignCommon(local, devicePath);
    if (device->common == NULL)
    {
        xf86Msg(X_INFO, "%s: '%s': No common pointer.\n",
                methodName,
                dev->identifier);
        goto SetupProc_fail;
    }
    /* Process the common options
*/
    xf86ProcessCommonOptions(local, local->options);

    /* Optional configuration items, now..
*/
    xf86Msg(X_CONFIG, "%s: '%s' path='%s'\n",
            methodName,
            dev->identifier,
            devicePath);

    /* DebugLevel
*/
    device->debugLevel = xf86SetIntOption(local->options, "DebugLevel",      INI_DEBUG_LEVEL);

    if (device->debugLevel > 0)
    {
        xf86Msg(X_CONFIG, "%s: Debug level set to %d\n",
                methodName, device->debugLevel);
    }

    /* HistorySize
*/
    local->history_size = xf86SetIntOption(local->options, "HistorySize", 0);
    if (local->history_size > 0)
    {
        xf86Msg(X_CONFIG, "%s: History size set to %d\n",
                methodName, local->history_size);
    }

    /* ScreenNo
*/
    device->screenNo = xf86SetIntOption(local->options, "ScreenNo",
                                        VALUE_NA);
    if (device->screenNo != VALUE_NA)
    {
        xf86Msg(X_CONFIG, "%s: '%s': attached to screen number %d\n",
                methodName,
                dev->identifier,
                device->screenNo);
    }

    /* AlwaysCore
*/
    if (xf86SetBoolOption(local->options, "AlwaysCore", FALSE) == TRUE)
    {
        device->flags |= ALWAYS_CORE_FLAG;
        xf86Msg(X_CONFIG, "%s: '%s': always core\n",
                methodName,
                dev->identifier);
    }

    device->useHalInput = FALSE;
    if (xf86SetBoolOption(local->options, "UseHal", FALSE) == TRUE)
    {
        device->useHalInput = TRUE;;
        xf86Msg(X_CONFIG, "%s: Use HAL True\n",methodName);
    }

    if (!device->useHalInput)
    {
        char *useHal = xf86SetStrOption(local->options, "_source", "");

        xf86Msg(X_CONFIG, "%s: _source: '%s'\n",methodName,useHal?useHal:"NULL");

        if (!strcmp(useHal, "server/hal"))
        {
            device->useHalInput = TRUE;;
            xf86Msg(X_CONFIG, "%s: Use HAL True\n",methodName);
        }
    }



    device->flags |= ABSOLUTE_FLAG;

    /* Cursor
*/
    device->flags |= CURSOR_STYLUS_FLAG;

    xf86Msg(X_CONFIG, "%s: %s cursor is in %s mode\n",
            methodName,
            local->name,
            (device->flags & CURSOR_STYLUS_FLAG) != 0 ? "stylus" : "relative");

#ifdef LINUX_INPUT
    /* The define-name is accurate; the XFree86 keyword is not. We are
* reading from a Linux kernel "Input" device. The Input device
* layer generally supports mice, joysticks, and keyboards. As
* an extension, the Input device layer also supports HID devices.
* HID is a standard specified by the USB Implementors Forum. Ergo,
* 99.9% of HID devices are USB devices.
*
* This option is misnamed, misunderstood, misanthrope. Maybe.
*/
    if (xf86SetBoolOption(local->options, "USB", TRUE) == TRUE)
    {
        xf86Msg(X_CONFIG, "%s: %s: USB: True\n",
                methodName,
                dev->identifier);
    }
#else
#error(The Promethean USB Driver is not available for your platform);
#endif

    /* If we've created a new common struct, it's unopened. So let's
* do a kludge, quietly open the tablet, and gather it's physical
* capacites...
*/
    /*
if (device->common->fd             == VALUE_NA &&
PrometheanGetTabletCapacity(local) == BadMatch)
{
xf86Msg(X_INFO, "%s: abends: PrometheanGetTabletCapacity fails\n",
methodName);
goto SetupProc_fail;
}*/

    /* ... With said capacities, we can tune the active area parameters
* for their sanity, which we'll need done before we later call
* PrometheanAllocateDriverAxes().
*/
    if (PrometheanNormalizeSizeParameters(local) == BadMatch)
    {
        xf86Msg(X_INFO, "%s: abends: PrometheanNormalizeSizeParameters fails\n",
                methodName);
        goto SetupProc_fail;
    }

    xf86Msg(X_CONFIG, "%s: ends successfully for %s\n",
            methodName, dev->identifier);

    /* Mark the device as configured
*/
    local->flags |= (XI86_POINTER_CAPABLE | XI86_CONFIGURED);

    /* Return the LocalDevice
*/
    return local;

    SetupProc_fail:
    /* Else we have to do a clean-up.
*/
    if (local != NULL)
    {
        device = (PrometheanDevicePtr)local->private;
        if (device != NULL)
        {
            PrometheanDetachCommon(local);
            xfree(device);
            /* do not set device to null, set the local object */
            local->private = NULL;
        }
        xfree(local);
        local = NULL;
    }
    return NULL;
}

/***********************************************************************
* PrometheanAllocateLocal --
*
* Allocates the device structures for the Promethean pseudo-driver instance.
*/
static LocalDevicePtr
        PrometheanAllocateLocal(
                InputDriverPtr drv,
                char*          name,
                char*          devicePath,
                int            deviceId,
                int            flags)
{
    const char*       methodName = "PrometheanAllocateLocal";
    PrometheanDevicePtr   device;
    LocalDevicePtr    local;

    CHECK_INPUT_DRIVER_PTR(drv, __LINE__);

    xf86Msg(X_INFO, "%s: begins with %s (%s, flags %d)\n",
            methodName, name, devicePath, flags);

    device = (PrometheanDevicePtr) xalloc(sizeof(PrometheanDeviceRec));
    if (device == NULL)
    {
        xf86Msg(X_INFO, "%s: failed to allocate PrometheanDevicePtr\n",
                methodName);
        return NULL;
    }

    local = xf86AllocateInput(drv, 0);
    if (local == NULL)
    {
        xf86Msg(X_INFO, "%s: failed at xf86AllocateInput()\n",
                methodName);
        xfree(device);
        return NULL;
    }

    local->name                     = name;
    local->type_name                = XI_TABLET;
    local->flags                    = (XI86_POINTER_CAPABLE | XI86_CONFIGURED);
    local->device_control           = PrometheanDispatcher;
    local->read_input               = PrometheanReadInput;
    local->control_proc             = PrometheanChangeControl;
    local->close_proc               = PrometheanCloseDriver;
    local->switch_mode              = PrometheanSwitchMode;
    local->conversion_proc          = PrometheanConvert;
    local->reverse_conversion_proc  = PrometheanReverseConvert;

    local->fd                       = VALUE_NA;
    local->atom                     = 0;
    local->dev                      = NULL;
    local->private                  = device;
    local->private_flags            = 0;
    local->history_size             = 0;
    local->old_x                    = VALUE_NA;
    local->old_y                    = VALUE_NA;

    /* Of note: devicePath is a deep copy; PrometheanCommonPtr is NULL.
*/
    PrometheanInitializeDevicePtr(device, xstrdup(devicePath),
                                  deviceId, flags, NULL);

    xf86Msg(X_INFO, "%s: ends successfully\n", methodName);
    return local;
}

/***********************************************************************
* PrometheanAssignCommon --
*
* Finds a common structure for this Promethean Driver instance that's been
* assigned to another Promethean Driver instance (yet for the same physical
* tablet.) If none exist, we'll create a new common struct for the user.
*/
static PrometheanCommonPtr
        PrometheanAssignCommon(
                LocalDevicePtr      local,
                char*               devicePath)
{
    PrometheanDevicePtr device = (PrometheanDevicePtr) local->private;
    PrometheanCommonPtr common;
    LocalDevicePtr  localIter;
    int i;

    CHECK_LOCAL_DEVICE_PTR(local, __LINE__);
    CHECK_PROMETHEAN_DEVICE_PTR(device, __LINE__);

    /* Go through linked list of local device drivers and find instances
* of the promethean driver. For those that refer to the same driverPath,
* get the common structure and return it.
*
* If said common structure does not yet exist, we'll proceed to
* create one.
*/
    for (localIter  = xf86FirstLocalDevice();
    localIter != NULL;
    localIter  = localIter->next)
    {
        common = (localIter->device_control == PrometheanDispatcher) ?
                 ((PrometheanDevicePtr)localIter->private)->common : NULL;

        if (local                           != localIter &&
            localIter->device_control       == PrometheanDispatcher &&
            common                          != NULL &&
            common->devicePath              != NULL &&
            strcmp(common->devicePath, devicePath) == 0)
        {
            /* Return the common with it having a reference back to local.
*/
            for (i = 0; i < PROMETHEAN_MAX_DEVICES; ++i)
            {
                if (common->deviceArray[i] == NULL)
                {
                    common->deviceArray[i] = local;
                    device->common = common;
                    break;
                }
            }
            return common;
        }
    }

    return PrometheanAllocateCommon(local, devicePath);
}

/***********************************************************************
* PrometheanAllocateCommon -
*
* This routine creates a 'common info' struct. There's only one
* per physical device path.
*
* This routine will also establish a reference to it's local struct.
*
* DO NOT CALL THIS METHOD DIRECTLY; go through PrometheanAssignCommon.
*/
static PrometheanCommonPtr
        PrometheanAllocateCommon(
                LocalDevicePtr      local,
                char*               devicePath)
{
    const char*     methodName = "PrometheanAllocateCommon";
    PrometheanDevicePtr device = (PrometheanDevicePtr) local->private;
    PrometheanCommonPtr common;

    CHECK_LOCAL_DEVICE_PTR(local, __LINE__);
    CHECK_PROMETHEAN_DEVICE_PTR(device, __LINE__);

    xf86Msg(X_INFO, "%s: begins\n", methodName);

    common = (PrometheanCommonPtr) xalloc(sizeof(PrometheanCommonRec));
    if (common == NULL)
    {
        xf86Msg(X_INFO, "%s: failed to allocate PrometheanCommonPtr\n",
                methodName);
        return NULL;
    }

    /* Record of the event currently being read of the queue
*/
    common->currentValues.eventType = 0;   /* Device event is for, e.g., */
    /* STYLUS, RUBBER, CURSOR */
    common->currentValues.x         = 0;   /* X coordinate */
    common->currentValues.y         = 0;   /* Y coordinate */
    common->currentValues.z         = 0;   /* Z (pressure) */
    common->currentValues.xTilt     = 0;   /* XTilt */
    common->currentValues.yTilt     = 0;   /* YTilt */
    common->currentValues.proximity = 0;   /* proximity bit */
    common->currentValues.macroKey  = VALUE_NA;   /* tablet macro key code */
    common->currentValues.button    = 0;   /* bitmask of buttons pressed */
    common->currentValues.wheel     = 0;   /* wheel */
    common->currentValues.distance  = 0;   /* currently unsupported */

    /* Record of the event previously read off of the queue. It's a
* dup of currentValues.
*/
    memmove(&common->previousValues, &common->currentValues,
            sizeof(PrometheanStateRec));

    common->devicePath              = xstrdup(devicePath);
    /* kernel driver path */
    common->flags                   = 0;             /* various flags */
    common->fd                      = VALUE_NA;      /* common fd */
    common->deviceArray[0]          = local;         /* local as element */
    common->deviceArray[1]          = NULL;          /* zap other elements */
    common->deviceArray[2]          = NULL;          /* zap other elements */
    common->xMinCapacity            = VALUE_NA;      /* tablet's min X value */
    common->xMaxCapacity            = VALUE_NA;      /* tablet's max X value */
    common->yMinCapacity            = VALUE_NA;      /* tablet's min Y value */
    common->yMaxCapacity            = VALUE_NA;      /* tablet's max Y value */
    common->zMinCapacity            = VALUE_NA;      /* tablet's min Z value */
    common->zMaxCapacity            = VALUE_NA;      /* tablet's max Z value */
    common->xtMinCapacity           = VALUE_NA;      /* tablet's min Xtilt */
    common->xtMaxCapacity           = VALUE_NA;      /* tablet's max XTilt */
    common->ytMinCapacity           = VALUE_NA;      /* tablet's min YTilt */
    common->ytMaxCapacity           = VALUE_NA;      /* tablet's max YTilt */
    common->wMinCapacity            = VALUE_NA;      /* tablet's min Wheel */
    common->wMaxCapacity            = VALUE_NA;      /* tablet's max Wheel */

    device->common                  = common;        /* set up info ptr */

    xf86Msg(X_INFO, "%s: ends\n", methodName);

    return common;
}

/***********************************************************************
* PrometheanDetachCommonPtr --
*
* Remove the common ptr reference from any closed devices.
*/
static Bool
        PrometheanDetachCommon(
                LocalDevicePtr local)
{
    PrometheanDevicePtr device = (PrometheanDevicePtr) local->private;
    PrometheanCommonPtr common = device->common;
    int i, j;

    CHECK_LOCAL_DEVICE_PTR(local, __LINE__);
    CHECK_PROMETHEAN_DEVICE_PTR(device, __LINE__);

    /* Paranoia is good :-)
*/
    if (common == NULL)
    {
        return 0;
    }

    /* Go through all devices referenced by the common struct.
* 1. When a device is closed, it should no longer reference
*    a common struct.
* 2. The device struct should have no reference to common.
* 3. When there no longer are devices that reference the
*    common struct, we need to delete the common struct.
*/
    for (i = 0, j=0; i < PROMETHEAN_MAX_DEVICES; ++i)
    {
        /* Remove the association
*/
        if (common->deviceArray[i] == local)
        {
            common->deviceArray[i] = NULL;
            device->common = NULL;
        }
        /* Now count the existing references.
*/
        if (common->deviceArray[i] != NULL)
        {
            ++j;
        }
    }
    /* If there are no more references, get rid of the common struct.
*/
    if (j == 0)
    {
        if (common->devicePath != NULL)
        {

            xfree(common->devicePath);
            common->devicePath = NULL;
        }
        xfree(common);
        common = NULL;
    }
    return j;
}

/***********************************************************************
* PrometheanInitializeDevicePtr
*
* Initializes the PrometheanDevicePtr to known 'N/A' values. Except for,
* of course, devicePath, deviceId, flags and common. Broken out as a
* separate routine as we call this also from the device driver callback
* when reprogramming the driver's state.
*/
static void
        PrometheanInitializeDevicePtr(
                PrometheanDevicePtr     device,
                char*               devicePath,
                int                 deviceId,
                int                 flags,
                PrometheanCommonPtr     common)
{
    device->devicePath  = devicePath;   /* kernel's driver path */
    device->deviceId    = deviceId;     /* device type      */
    device->flags       = flags;        /* flags for various settings */
    device->xSize       = VALUE_NA;     /* Active Area X */
    device->ySize       = VALUE_NA;     /* Active Area Y */
    device->xOffset     = VALUE_NA;     /* Active area offset X */
    device->yOffset     = VALUE_NA;     /* Active area offset Y */
    device->xMax        = VALUE_NA;     /* Max allowed X value */
    device->yMax        = VALUE_NA;     /* Max allowed Y value */
    device->zMin        = VALUE_NA;     /* Min allowed Z value */
    device->zMax        = VALUE_NA;     /* Max allowed Z value */
    device->xTop        = VALUE_NA;     /* Upper Left X coordinate */
    device->yTop        = VALUE_NA;     /* Upper Left Y coordinate */
    device->xBottom     = VALUE_NA;     /* Lower Right X coordinate */
    device->yBottom     = VALUE_NA;     /* Lower Right Y coordinate */
    device->xThreshold  = VALUE_NA;     /* X threshold */
    device->yThreshold  = VALUE_NA;     /* Y threshold */
    device->zThreshold  = VALUE_NA;     /* Z threshold  */
    device->zMode       = VALUE_NA;     /* Z: linear, soft, hard log */
    device->screenNo    = VALUE_NA;     /* Attached to X screen */
    device->common      = common;       /* The commonPtr        */
    device->message.messageBuffer = NULL;

}

/***********************************************************************
* PrometheanOpenDriver --
*
* Opens the file descriptor associated with the Promethean driver;
* also performs axis allocations and normalization of drawing area
* parameters.
*
* This should be the only 'public' method invoked to open an Promethean
* Device Driver.
*/
static Bool
        PrometheanOpenDriver(
                DeviceIntPtr dev)
{
    const char*     methodName = "PrometheanOpenDriver";
    LocalDevicePtr  local  = (LocalDevicePtr) dev->public.devicePrivate;
    PrometheanDevicePtr device = (PrometheanDevicePtr) PRIVATE(dev);
    device->devPtr = dev;

    CHECK_DEVICE_INT_PTR(dev, __LINE__);
    CHECK_LOCAL_DEVICE_PTR(local, __LINE__);
    CHECK_PROMETHEAN_DEVICE_PTR(device, __LINE__);

    DBG(10, xf86Msg(X_INFO, "%s: begins (%s)\n",
                    methodName,
                    local->name));

    if (PrometheanOpenFileDescriptor(local)      == BadMatch ||
        PrometheanNormalizeSizeParameters(local) == BadMatch ||
        PrometheanAllocateDriverAxes(dev)        == BadMatch)
    {
        xf86Msg(X_INFO, "%s: abends: failure opening/allocating driver\n",
                methodName);
        return BadMatch;
    }

    DBG(10, xf86Msg(X_INFO, "%s: ends successfully (%s)\n",
                    methodName,
                    local->name));

    return Success;
}

/***********************************************************************
* PrometheanOpenFileDescriptor
*
* Opens the UNIX file descriptor associated with an Promethean Driver instance.
* Deals with the issue of commonality of Promethean Drivers' file descriptors
* to their physical Promethean tablet (e.g., the drivers all share the same
* fd, a copy of which is kept in the common struct.)
*
* DO NOT CALL THIS ROUTINE DIRECTLY, with the exception of
* from PrometheanOpenDriver().
*/
static Bool
        PrometheanOpenFileDescriptor(
                LocalDevicePtr local)
{
    const char*     methodName = "PrometheanOpenFileDescriptor";
    PrometheanDevicePtr device = (PrometheanDevicePtr) local->private;
    PrometheanCommonPtr common = device->common;
    int i;

    DBG(10, xf86Msg(X_INFO, "%s: begins (%s)\n",
                    methodName,
                    local->name));

    CHECK_LOCAL_DEVICE_PTR(local, __LINE__);
    CHECK_PROMETHEAN_DEVICE_PTR(device, __LINE__);
    CHECK_PROMETHEAN_COMMON_PTR(common, __LINE__);


    /* We're being asked to open a driver that already has an opened
* file handle? Must be a redundency; play along.
*/
    if (PrometheanIsValidFileDescriptor(local->fd) == Success)
    {
        DBG(10, xf86Msg(X_INFO, "%s: ends successfully (%s,fd=%d)\n",
                        methodName,
                        local->name,
                        local->fd));
        return Success;
    }

    /* If we are the first device being opened for a physical tablet,
* then we'll have no common struct assign. So, let's build one.
*/
    if (common == NULL)
    {
        if ((common = PrometheanAssignCommon(local, device->devicePath)) == NULL)
        {
            xf86Msg(X_INFO, "%s: abends on PrometheanAssignCommon\n",
                    methodName);
            return BadMatch;
        }
        device->common = common;
    }

    /* The common struct will have a valid fd if one of the other Promethean
* driver instances associated with the physical tablet have already
* been Open'ed. If so, we'll reuse that fd. Else, we're the first
* Promethean driver, and we have to do the opening.
*/
    if (PrometheanIsValidFileDescriptor(common->fd) == Success)
    {
        local->fd = common->fd;
        DBG(10, xf86Msg(X_INFO, "%s: ends successfully (%s,fd=%d)\n",
                        methodName,
                        local->name,
                        local->fd));
        return Success;
    }

    /* Okay, we've done everything we could to prevent an actual UNIX
* 'open' call being invoked. We've now no choice :-)
*/
    if (PrometheanLowLevelOpen(local) != Success)
    {
        xf86Msg(X_INFO, "%s: abends on PrometheanLowLevelOpen\n",
                methodName);
        return BadMatch;
    }

    /* Move the file descriptor over to the common struct to prevent
* subsequent opens of the physical tablet by other Promethean driver
* instances.
*/
    if (PrometheanIsValidFileDescriptor(common->fd) == BadMatch)
    {
        common->fd = local->fd;
    }

    /* Copy the opened file descriptor over to any other Promethean Driver
* instances who share the same physical device, and have already
* been assigned to this common struct.
*/
    for (i = 0; i < PROMETHEAN_MAX_DEVICES; ++i)
    {
        if (common->deviceArray[i] != NULL)
        {
            common->deviceArray[i]->fd     = local->fd;
            common->deviceArray[i]->flags |=
                    (XI86_CONFIGURED | XI86_POINTER_CAPABLE);
        }
    }

    DBG(10, xf86Msg(X_INFO, "%s: ends successfully (%s,fd=%d)\n",
                    methodName,
                    local->name,
                    local->fd ));

    return Success;
}

/*
* Multiplex all devices into a virtual device for xorg not 7.4.
* Problem is if xinput and hal are not enabled in 7.4, the driver will stop working.
* */
void *threadReader(void *p)
{
    const char * methodName = "threadPrometheanReader";
    char *path = (char*)p;
    int ret = 0;
    int fr = open(path,O_RDONLY);
    int fw = open("/dev/input/activdevices",O_WRONLY);


    xf86Msg(X_INFO, "%s: read %d, write %d, path %s\n",
            methodName,
            fr,
            fw,
            path);

    if ((fr >=0) && (fw >=0))
    {
        int sz = 0;
        unsigned char buffer[256];
        unsigned char loop = 1;

        SYSCALL(ret, ioctl(fr, EVIOCGRAB, (pointer)1));
        if (ret < 0) {
            xf86Msg(X_WARNING, "%s can't grab event device: error %d\n",	path, errno);
        }

        while (loop)
        {
            sz = read(fr,buffer,sizeof(buffer));
            loop = sz >= 0;

            if (!loop)
            {
                switch (errno)
                {
                case EINTR: /* interrupted by a signal. It should be ok to continue. */
                    loop = 1;
                    xf86Msg(X_INFO, "%s: Reader interrupted but continue: errno %d %08x %s\n",
                            methodName,
                            errno,
                            errno,
                            strerror(errno));
                    break;
                default:
                    xf86Msg(X_INFO, "%s: Reader exit: errno %d %08x %s\n",
                            methodName,
                            errno,
                            errno,
                            strerror(errno));
                    break;
                }
            }

            if (sz > 0)
            {
                /* Do we really need a crit sect ? At the end of the day, if you move two mice at the same time,
     * the pointer will be jumping. If it's jumping because it gets two coordinates at different
     * position or because it gets two packets intermixed together, at the end of the day,
     * it's jumping just the same ! */
                if (write(fw,buffer,sz) != sz)
                    xf86Msg(X_WARNING, "%s Error writing to multiplexer: error %d\n",	path, errno);
            }
        }

        SYSCALL(ret, ioctl(fr, EVIOCGRAB, (pointer)0));
        if (ret < 0) {
            xf86Msg(X_WARNING, "%s: can't ungrab event device %s: error %d\n",methodName,	path, errno);
        }
    }


    if (fr >= 0)
        close(fr);

    if (fw >= 0)
        close(fw);
    free(path);
    xf86Msg(X_INFO, "%s: Reader exit\n",
            methodName);
    return NULL;
}

void *threadDeviceMonitor(void* p)
{
    pthread_t h;
    char *path;

    int d,newd;
    XF86DIR *dir;
    struct {
        ino_t inode;
        char path[256];
    } stats[PROMETHEAN_MAX_DEVICES], newstats[PROMETHEAN_MAX_DEVICES];

    struct stat stbuf;
    const char *methodName={"threadDeviceMonitorPromethean"};

    xf86Msg(X_INFO, "%s: start\n",	methodName);

    memset(stats,0,sizeof(stats));

    while (1)
    {
        memset(newstats,0,sizeof(newstats));
        dir = xf86opendir("/dev/input");
        if (dir)
        {
            XF86DIRENT *de = xf86readdir(dir);
            int total = 0;

            while (de)
            {
                if (!strncasecmp(de->d_name,"promethean_",strlen("promethean_")))
                {
                    /* check the stat */
                    sprintf(newstats[total].path,"/dev/input/%s",de->d_name);

                    stat(newstats[total].path,&stbuf);
                    newstats[total].inode = stbuf.st_ino;
                    total++;

                    if (total >= PROMETHEAN_MAX_DEVICES)
                    {
                        xf86Msg(X_INFO, "Promethean : too many devices for X.");
                        break;
                    }
                }

                de = xf86readdir(dir);
            }

            closedir(dir);

            for (newd = 0; newd < total; newd++)
            {

                for (d=0; d < PROMETHEAN_MAX_DEVICES; d++)
                {
                    if (stats[d].inode == newstats[newd].inode)
                        break;
                }


                if (d >= PROMETHEAN_MAX_DEVICES) /* device not found, add it to the array */
                {

                    path = strdup(newstats[newd].path);
                    if (path)
                    {
                        xf86Msg(X_INFO, "%s: found new device %s : inode %d\n",	methodName, path, (int)newstats[newd].inode);
                        pthread_create(&h,NULL,threadReader,path);
                    }
                }
            }

            memcpy(stats,newstats,sizeof(stats));
        }

        /* latency of 3 seconds before a device is picked up. It's sufficient. */
        sleep(3);
    }
}


/***********************************************************************
* PrometheanLowLevelOpen --
*
* This is the low-level call to open the file descriptor associated
* with an Promethean Device Driver instance. Not only do we open the
* file descriptor, but we also perform the ioctl commands to learn
* the physical tablet's capacities, which we store in the common
* struct.
*
* DO NOT CALL THIS ROUTINE DIRECTLY -- use PrometheanOpenFileDescriptor.
* EXCEPTION: PrometheanGetTabletCapacity().
*/
static Bool
        PrometheanLowLevelOpen(LocalDevicePtr local)
{
    const char*     methodName = "PrometheanLowLevelOpen";
    static pthread_t h = (pthread_t)NULL;

    PrometheanDevicePtr device = (PrometheanDevicePtr)local->private;
    PrometheanCommonPtr common = device->common;


    DBG(10, xf86Msg(X_INFO, "%s: begins, (%s,path=%s)\n",
                    methodName,
                    local->name,
                    common->devicePath));

    CHECK_LOCAL_DEVICE_PTR(local, __LINE__);
    CHECK_PROMETHEAN_DEVICE_PTR(device, __LINE__);
    CHECK_PROMETHEAN_COMMON_PTR(common, __LINE__);

    /* Do the low-level open(2) call.
*/
    local->fd = xf86OpenSerial(local->options);
    if (local->fd < 0)
    {
        xf86Msg(X_INFO, "%s: device not present: (%s,path=%s): %s\n",
                methodName,
                local->name,
                common->devicePath,
                strerror(errno));
        local->fd  = VALUE_NA;
        common->fd = VALUE_NA;
        /* return BadMatch; */
        return Success;
    }
    /* this is constant across all devices (the kernel driver always output his range). */
    common->xMinCapacity = 0;
    common->xMaxCapacity = 32767;

    common->yMinCapacity = 0;
    common->yMaxCapacity = 32767;

    common->zMinCapacity = 0;
    common->zMaxCapacity = 1024;


    /* start thread to monitor pnp. */
    if (device->useHalInput)
    {
        int ret = 0;

        SYSCALL(ret, ioctl(local->fd, EVIOCGRAB, (pointer)1));
        if (ret < 0) {
            xf86Msg(X_WARNING, "%s: can't grab event device %s: error %d\n",methodName,	common->devicePath, errno);
        }

        DBG(2, xf86Msg(X_INFO, "%s: Using HAL input, no need to monitor Promethean devices.\n",
                       methodName));

    }
    else
    {
        DBG(2, xf86Msg(X_INFO, "%s: Starting Monitor for Promethean devices.\n",methodName));
        if (h == (pthread_t)NULL)
            pthread_create(&h,NULL,threadDeviceMonitor,NULL);
    }

    DBG(2, xf86Msg(X_INFO, "%s: ends successfully (%s,path=%s)\n",
                   methodName,
                   local->name,
                   common->devicePath));


    return Success;
}


/***********************************************************************
* PrometheanNormalizeSizeParameters --
*
* Normalize active area parameters so they actually fit on the tablet.
*/
static Bool PrometheanNormalizeSizeParameters(LocalDevicePtr local)
{
    const char*     methodName = "PrometheanNormalizeSizeParameters";
    PrometheanDevicePtr device = (PrometheanDevicePtr) local->private;
    PrometheanCommonPtr common = device->common;
    double          tabletRatio, screenRatio, xFactor, yFactor;
    int             gap, activeAreaType, message, xDiff, yDiff;

    CHECK_LOCAL_DEVICE_PTR(local, __LINE__);
    CHECK_PROMETHEAN_DEVICE_PTR(device, __LINE__);
    CHECK_PROMETHEAN_COMMON_PTR(common, __LINE__);

    DBG(20, xf86Msg(X_INFO, "%s: begins\n", methodName));

    /* if the device is not present, just return success */
    /*if (local->fd < 0)
return Success;*/


    /* Check our active area parameters. We support the following
* three sets of mutually exclusive parameter sets:
*
* 1) xMax/yMax. The origin {0,0} of the active area is the origin
*    of the physical tablet. You therefore are describing the
*    width and height of that active area.
*
* 2) xOffset/xSize,yOffset/ySize. The origin (0,0) of the active
*    area is defined as (xOffset,yOffset) (which we'll report as
*    {0,0}). The size of the active area in width and height are
*    expressed in coordinates as xSize/ySize. That is to say,
*    if xOffset=5; yOffset=5, and xSize=10; ySize=10, then we will
*    have an active area beginning at {5,5} and ending at {15,15}.
*    Physical coordinate {5,5} is reported as {0,0}; {15,15} is
*    reported as {10,10}. The rest of the tablet is inert, as far as
*    drawing area goes,
*
* 3) xTop/xBottom,yTop/yBottom. The difference between this and
*    #2 above is that all four parameters are physical coordinates
*    on the tablet. Using the example above, xTop=5; yTop=5, and
*    xBottom=15; yBottom=15. It is inferred mathematically that
*    the overall active area is 10x10 coordinates.
*/
    activeAreaType = PROMETHEAN_ACTIVE_AREA_NONE;
    if (device->xMax != VALUE_NA ||
        device->yMax != VALUE_NA)
    {
        activeAreaType |= PROMETHEAN_ACTIVE_AREA_MAX;
    }
    if (device->xSize   != VALUE_NA ||
        device->ySize   != VALUE_NA ||
        device->xOffset != VALUE_NA ||
        device->yOffset != VALUE_NA)
    {
        activeAreaType |= PROMETHEAN_ACTIVE_AREA_SIZE;
    }

    if (device->xTop    != VALUE_NA ||
        device->yTop    != VALUE_NA ||
        device->xBottom != VALUE_NA ||
        device->yBottom != VALUE_NA)
    {
        activeAreaType |= PROMETHEAN_ACTIVE_AREA_TOP;
    }

    /* Okay, we have a mask of which active area parameter(s)
* you've used by their type. What we now test is whether
* you have one and only one bit set.
*/
    if (activeAreaType != PROMETHEAN_ACTIVE_AREA_MAX &&
        activeAreaType != PROMETHEAN_ACTIVE_AREA_SIZE &&
        activeAreaType != PROMETHEAN_ACTIVE_AREA_TOP &&
        activeAreaType != PROMETHEAN_ACTIVE_AREA_NONE)
    {
        xf86Msg(X_INFO, "%s: abends: Mutually exclusive active area parms!\n",
                methodName);
        return BadMatch;
    }

    /* We're now going to deal with the active area parameters specifying
* an area larger than the actual tablet. We do this in three passes,
* as there are three sets of active area parameters the user could've
* used.
*
* The first of the bunch are the xMax/yMax set.
*/
    if (device->xMax != VALUE_NA ||
        device->yMax != VALUE_NA)
    {
        /* Deal with larger-than-tablet and NA values.
*/
        if (device->xMax >  common->xMaxCapacity ||
            device->xMax == VALUE_NA)
        {
            device->xMax = common->xMaxCapacity;
            xf86Msg(X_CONFIG, "%s: xMax value unspecified/invalid; now %d\n",
                    methodName,
                    device->xMax);
        }
        if (device->yMax > common->yMaxCapacity ||
            device->yMax == VALUE_NA)
        {
            device->yMax = common->yMaxCapacity;
            xf86Msg(X_CONFIG,"%s: yMax value unspecified/invalid; now %d\n",
                    methodName,
                    device->yMax);
        }

        /* Internally we use xTop/yTop/xBottom/yBottom
* for everything. It's the easiest for us to work
* with, vis-a-vis filtering.
*/
        device->xTop = 0;
        device->yTop = 0;
        device->xBottom = device->xMax;
        device->yBottom = device->yMax;
    }

    /* We're now doing the same 'active area greater than the tablet'
* tests as above, but now assuming the user has given us the
* xOffset/yOffset; xSize/ySize parameters.
*/
    if (device->xSize   != VALUE_NA ||
        device->ySize   != VALUE_NA ||
        device->xOffset != VALUE_NA ||
        device->yOffset != VALUE_NA)
    {
        message = 0;
        /* Simple sanity tests: nothing larger than the
* tablet; nothing negative, except for an NA value.
*/
        if (device->xOffset  != VALUE_NA &&
            (device->xOffset >  common->xMaxCapacity ||
             device->xOffset <  common->xMinCapacity))
        {
            message = 1;
            device->xOffset = 0;
        }
        if (device->yOffset  != VALUE_NA &&
            (device->yOffset >  common->yMaxCapacity ||
             device->yOffset <  common->yMinCapacity))
        {
            message = 1;
            device->yOffset = 0;
        }
        if (device->xSize  != VALUE_NA &&
            (device->xSize >  common->xMaxCapacity ||
             device->xSize <  common->yMinCapacity))
        {
            message = 1;
            device->xSize = common->xMaxCapacity;
        }
        if (device->ySize  != VALUE_NA &&
            (device->ySize >  common->yMaxCapacity ||
             device->ySize <  common->yMinCapacity))
        {
            message = 1;
            device->ySize = common->yMaxCapacity;
        }

        /* If one parameter is set but not the other, we'll
* guess at something reasonable for the missing one.
*/
        if (device->xOffset == VALUE_NA ||
            device->xSize   == VALUE_NA)
        {
            if (device->xOffset == VALUE_NA)
            {
                message = 1;
                device->xOffset = 0;
            }
            else
            {
                message = 1;
                device->xSize = common->xMaxCapacity - device->xOffset;
            }
        }
        if (device->yOffset == VALUE_NA ||
            device->ySize   == VALUE_NA)
        {
            if (device->yOffset == VALUE_NA)
            {
                message = 1;
                device->yOffset = 0;
            }
            else
            {
                message = 1;
                device->ySize = common->yMaxCapacity - device->yOffset;
            }
        }

        /* Do not allow the active area to exceed the size of the
* tablet. To do this, we have to consider both parameters.
* Assumption: xOffset/yOffset is always correct; deliver less
* of the tablet than they asked for, if the user asked for
* larger area than tablet can provide.
*/
        if ((device->xOffset + device->xSize) > common->xMaxCapacity)
        {
            message = 1;
            device->xSize = common->xMaxCapacity - device->xOffset;
        }
        if ((device->yOffset + device->ySize) > common->yMaxCapacity)
        {
            message = 1;
            device->ySize = common->yMaxCapacity - device->yOffset;
        }

        /* 'message' is used to indicate that we've changed some parameter
* during our filtration process. It's conceivable that we may
* have changed parameters several times, so we without commentary
* to the very end.
*/
        if (message == 1)
        {
            xf86Msg(X_CONFIG,
                    "%s: xOffset/yOffset: xSize/ySize are wrong/unspecified.\n",
                    methodName);
            xf86Msg(X_CONFIG,"%s: xOffset now %d\n",
                    methodName, device->xOffset);
            xf86Msg(X_CONFIG,"%s: yOffset now %d\n",
                    methodName, device->yOffset);
            xf86Msg(X_CONFIG,"%s: xSize now %d\n",
                    methodName, device->xSize);
            xf86Msg(X_CONFIG,"%s: ySize now %d\n",
                    methodName, device->ySize);
        }

        /* Internally we use xTop/yTop/xBottom/yBottom
* for everything. It's the easiest for us to work
* with, vis-a-vis filtering.
*/
        device->xTop    = device->xOffset;
        device->yTop    = device->yOffset;
        device->xBottom = device->xOffset + device->xSize;
        device->yBottom = device->yOffset + device->ySize;
    }

    /* We're now doing the same 'active area greater than the tablet'
* tests as above, but now assuming checking for the third set
* of parameters the user can specify, the xTop/yTop/xBottom/yBottom
* parameters.
*
* As all tablet size parameters are internally expressed as xTop/yTop,
* etc., we'll do our 'greater than tablet' tests on the xTop/yTop
* parameters we synthesized from the above tests as well. So, these
* tests are being done NO MATTER WHAT 'size' parameters you actually
* specified.
*/
    if (device->xTop == VALUE_NA ||
        device->xTop <  common->xMinCapacity ||
        device->xTop >  common->xMaxCapacity)
    {
        device->xTop = 0;
        xf86Msg(X_CONFIG,"%s: xTop invalid/unspecified: now %d\n",
                methodName, device->xTop);
    }
    if (device->yTop == VALUE_NA ||
        device->yTop <  common->yMinCapacity ||
        device->yTop >  common->yMaxCapacity)
    {
        device->yTop = 0;
        xf86Msg(X_CONFIG,"%s: yTop invalid/unspecified: now %d\n",
                methodName, device->yTop);
    }
    if (device->xBottom == VALUE_NA ||
        device->xBottom <  common->xMinCapacity ||
        device->xBottom >  common->xMaxCapacity)
    {
        device->xBottom = common->xMaxCapacity;
        xf86Msg(X_CONFIG,"%s: xBottom invalid/unspecified: now %d\n",
                methodName, device->xBottom);
    }
    if (device->yBottom == VALUE_NA ||
        device->yBottom <  common->yMinCapacity ||
        device->yBottom >  common->yMaxCapacity)
    {
        device->yBottom = common->yMaxCapacity;
        xf86Msg(X_CONFIG,"%s: yBottom invalid/unspecified: now %d\n",
                methodName, device->yBottom);
    }

    /* Determine the X screen we're going to be using.
* If NA, or larger than the number of screens we
* have, or negative, we've going for screen 0, e.g.,
* 'default' screen.
*/
    if ( device->screenNo >= screenInfo.numScreens ||
         device->screenNo == VALUE_NA ||
         device->screenNo < 0)
    {
        device->screenNo = 0;
        xf86Msg(X_CONFIG,"%s: ScreenNo invalid/unknown; now %d\n",
                methodName, device->screenNo);
    }

    /* Calculate the ratio according to KeepShape, TopX and TopY
*/
    if ((device->flags & KEEP_SHAPE_FLAG) != 0)
    {
        xDiff = common->xMaxCapacity - device->xTop;
        yDiff = common->yMaxCapacity - device->yTop;

        tabletRatio = (double) xDiff / (double) yDiff;
        screenRatio = (double) screenInfo.screens[device->screenNo]->width /
                      (double) screenInfo.screens[device->screenNo]->height;

        DBG(2, xf86Msg(X_INFO,
                       "%s: Screen %d: screenRatio = %.3g, tabletRatio = %.3g\n",
                       methodName, device->screenNo,
                       screenRatio, tabletRatio));

        if (screenRatio > tabletRatio)
        {
            gap = (int)((double)common->yMaxCapacity *
                        (1.0 - tabletRatio/screenRatio));
            device->xBottom = common->xMaxCapacity;
            device->yBottom = common->yMaxCapacity - gap;
            DBG(20, xf86Msg(X_INFO, "%s: Screen %d: 'Y' Gap of %d computed\n",
                            methodName, device->screenNo, gap));
        }
        else
        {
            gap = (int)((double)common->xMaxCapacity *
                        (1.0 - screenRatio/tabletRatio));
            device->xBottom = common->xMaxCapacity - gap;
            device->yBottom = common->yMaxCapacity;
            DBG(20, xf86Msg(X_INFO, "%s: Screen %d: 'X' Gap of %d computed\n",
                            methodName, device->screenNo, gap));
        }
    }

    xFactor = (double)screenInfo.screens[device->screenNo]->width/
              (double)(device->xBottom - device->xTop);
    yFactor = (double)screenInfo.screens[device->screenNo]->height/
              (double)(device->yBottom - device->yTop);

    /* Check threshold correctness
*/
    if (device->xThreshold > common->xMaxCapacity ||
        device->xThreshold < common->xMinCapacity ||
        device->xThreshold == VALUE_NA)
    {
        device->xThreshold = 0;
    }

    if (device->yThreshold > common->yMaxCapacity ||
        device->yThreshold < common->yMinCapacity ||
        device->yThreshold == VALUE_NA)
    {
        device->yThreshold = 0;
    }

    if (device->zThreshold > common->zMaxCapacity ||
        device->zThreshold < common->zMinCapacity ||
        device->zThreshold == VALUE_NA)
    {
        device->zThreshold = 0;
    }

    if (device->zMin < common->zMinCapacity ||
        device->zMin == VALUE_NA)
    {
        device->zMin = common->zMinCapacity;
    }

    if (device->zMax > common->zMaxCapacity ||
        device->zMax == VALUE_NA)
    {
        device->zMax = common->zMaxCapacity;
    }

    DBG(20, xf86Msg(X_INFO, "%s: ends successfully\n", methodName));

    return Success;
}

/***********************************************************************
* PrometheanAllocateDriverStructs --
*
* This method is used to allocate Axis Structs and set up callbacks.
*/
static Bool
        PrometheanAllocateDriverAxes(
                DeviceIntPtr dev)
{
    const char* methodName = "PrometheanAllocateDriverAxes";
    LocalDevicePtr  local  = (LocalDevicePtr) dev->public.devicePrivate;
    PrometheanDevicePtr device = (PrometheanDevicePtr) local->private;
    PrometheanCommonPtr common = device->common;
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
    Atom *AxesLabels = NULL;
#endif    

#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
	AxesLabels = xcalloc(3, sizeof(ATOM));
    AxesLabels[0] = XIGetKnownProperty(AXIS_LABEL_PROP_ABS_X);
    AxesLabels[1] = XIGetKnownProperty(AXIS_LABEL_PROP_ABS_Y);
    AxesLabels[2] = XIGetKnownProperty(AXIS_LABEL_PROP_ABS_Z);
#endif
    CHECK_DEVICE_INT_PTR(dev, __LINE__);
    CHECK_LOCAL_DEVICE_PTR(local, __LINE__);
    CHECK_PROMETHEAN_DEVICE_PTR(device, __LINE__);
    CHECK_PROMETHEAN_COMMON_PTR(common, __LINE__);

    DBG(10, xf86Msg(X_INFO, "%s: begins\n", methodName));

    if (common == NULL)
    {
        xf86Msg(X_INFO, "%s: abends - No common struct\n", methodName);
        return BadMatch;
    }

    DBG(10, xf86Msg(X_INFO, "%s: X %d x %d\n", methodName,common->xMinCapacity,device->xBottom - device->xTop));
    DBG(10, xf86Msg(X_INFO, "%s: Y %d x %d\n", methodName,common->yMinCapacity,device->yBottom - device->yTop));

    /* Create axisStructs for every axis we support.
* NOTE: min_resolution and max_resolution infers to
* me a programmability to increase/decrease resolution.
* We don't support that, so min & max = current_resolution.
*/
    InitValuatorAxisStruct(dev,                     /* X resolution */
                           0,                                   /* axis_id */
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
                           AxesLabels[0],
#endif
                           common->xMinCapacity,                /* min value */
                           device->xBottom - device->xTop,      /* max value */
                           LPI2CPM(500),                        /* resolution */
                           0,                                   /* min_resolution */
                           LPI2CPM(500));                       /* max_resolution */

    InitValuatorAxisStruct(dev,                     /* Y Resolution */
                           1,                                   /* axis_id */
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
                           AxesLabels[1],
#endif
                           common->yMinCapacity,                /* min value */
                           device->yBottom - device->yTop,      /* max value */
                           LPI2CPM(500),                        /* resolution */
                           0,                                   /* min_resolution */
                           LPI2CPM(500));                       /* max_resolution */

    InitValuatorAxisStruct(dev,                     /* Pressure */
                           2,                                   /* axis_id */
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
                           AxesLabels[2],
#endif
                           common->zMinCapacity,                /* min value */
                           common->zMaxCapacity + 1,            /* max value */
                           1,                                   /* resolution */
                           1,                                   /* min resolution */
                           1);                                  /* max resolution */

    DBG(10, xf86Msg(X_INFO, "%s: ends successfully\n", methodName));

    return Success;
}


/***********************************************************************
* PrometheanCloseDriver --
*
* Close the device by disassociating the Promethean Driver instance from its
* common struct. If no Drivers are subsequently associated with the common
* struct, then we want to also do a low-level close() on the file descriptor.
*/
static void
        PrometheanCloseDriver(
                LocalDevicePtr local)
{
    const char* methodName = "PrometheanCloseDriver";
    PrometheanDevicePtr device = (PrometheanDevicePtr) local->private;

    CHECK_LOCAL_DEVICE_PTR(local, __LINE__);
    CHECK_PROMETHEAN_DEVICE_PTR(device, __LINE__);

    xf86Msg(X_INFO, "%s: begins\n", methodName);

    /* If we're asked to close a Driver that already has been
* closed (e.g., the file descriptor is invalid), assume it is
* a paranoid redundency, and let it pass without comment.
*/
    if (PrometheanIsValidFileDescriptor(local->fd) == Success)
    {
        /* Disassociate the Driver instance from it's common struct,
* returning the number of still-attached Drivers. Only if that
* number goes to zero, should we proceed to call the OS and
* close the file descriptor.
*/
        if (PrometheanDetachCommon(local) == 0)
        {
            if (device->useHalInput)
            {
                int ret = 0;

                SYSCALL(ret, ioctl(local->fd, EVIOCGRAB, (pointer)0));
                if (ret < 0) {
                    xf86Msg(X_WARNING, "%s: can't ungrab event device %s: error %d\n",methodName,	device->devicePath, errno);
                }
            }

            close(local->fd);
        }
    }
    local->fd = VALUE_NA;
    /*xfree(device->devicePath);
device->devicePath = NULL; */

    xf86Msg(X_INFO, "%s: ends\n", methodName);
}

/***********************************************************************
* PrometheanChangeControl --
*
* Allow the user to change the tablet resolution -- we have an issue
* insofar as we don't know how to write to the tablet. And furthermore,
* even if we DID know how to write to the tablet, it doesn't support
* a "change resolution" call. We tried to avoid this by claiming when
* creating axisStructs that minRes = curRes = maxRes. So, we should never
* get dispatched.
*/
static Bool
        PrometheanChangeControl(
                LocalDevicePtr local,
                xDeviceCtl     *control)
{
    const char*           methodName = "PrometheanChangeControl";
    PrometheanDevicePtr       device;
    xDeviceResolutionCtl *res;
    int                  *resolutions;

    device = (PrometheanDevicePtr) local->private;

    CHECK_LOCAL_DEVICE_PTR(local, __LINE__);
    CHECK_PROMETHEAN_DEVICE_PTR(device, __LINE__);

    DBG(10, xf86Msg(X_INFO, "%s: begins\n", methodName));

    res = (xDeviceResolutionCtl *)control;

    if ((control->control != DEVICE_RESOLUTION) ||
        (res->num_valuators < 1))
    {
        xf86Msg(X_INFO, "%s: abends\n", methodName);
        return BadMatch;
    }

    resolutions = (int *)(res +1);

    DBG(30, xf86Msg(X_INFO, "%s: changing to resolution %d\n",
                    methodName, resolutions[0]));

    DBG(10, xf86Msg(X_INFO, "%s: ends successfully\n", methodName));

    return Success;
}

/***********************************************************************
* PrometheanSwitchMode --
*
* Switches the mode between absolute or relative.
*/
static Bool
        PrometheanSwitchMode(
                ClientPtr       client,
                DeviceIntPtr    dev,
                int             mode)
{
    const char*     methodName = "PrometheanSwitchMode";
    LocalDevicePtr  local  = (LocalDevicePtr)dev->public.devicePrivate;
    PrometheanDevicePtr device = (PrometheanDevicePtr)(local->private);

    CHECK_DEVICE_INT_PTR(dev, __LINE__);
    CHECK_LOCAL_DEVICE_PTR(local, __LINE__);
    CHECK_PROMETHEAN_DEVICE_PTR(device, __LINE__);

    DBG(10, xf86Msg(X_INFO, "%s: begins: dev=0x%p mode=%d\n",
                    methodName, (void*) dev, mode));

    if (mode == Absolute)
    {
        device->flags |= ABSOLUTE_FLAG;
    }
    else if (mode == Relative)
    {
        device->flags &= ~ABSOLUTE_FLAG;
    }
    else
    {
        xf86Msg(X_INFO, "%s: dev=0x%p invalid mode=%d\n",
                methodName, (void*) dev, mode);
        return BadMatch;
    }

    DBG(10, xf86Msg(X_INFO, "%s: ends successfully\n", methodName));

    return Success;
}

/***********************************************************************
* PrometheanConvert
* Convert valuators to X and Y. We deal with multiple X screens, adjusting
* for xTop/xBottom/yTop/yBottom (or xSize/ySize).
*/
static Bool
        PrometheanConvert(
                LocalDevicePtr local,
                int            first,
                int            num,
                int            v0,
                int            v1,
                int            v2,
                int            v3,
                int            v4,
                int            v5,
                int*           x,
                int*           y)
{
    const char*     methodName = "PrometheanConvert";
    PrometheanDevicePtr device = (PrometheanDevicePtr) local->private;
    int             xSize, ySize;
    int             width, height;

    CHECK_LOCAL_DEVICE_PTR(local, __LINE__);
    CHECK_PROMETHEAN_DEVICE_PTR(device, __LINE__);

    DBG(10, xf86Msg(X_INFO, "%s: begins, with: first=%d, num=%d, v0=%d, "
                    "v1=%d, v2=%d, v3=%d,, v4=%d, v5=%d, x=%d, y=%d\n",
                    methodName, first, num, v0, v1, v2, v3, v4, v5, *x, *y));

    if (first != 0 || num == 1)
    {
        return BadMatch;
    }

    xSize = device->xBottom - device->xTop;
    ySize = device->yBottom - device->yTop;

    width  = screenInfo.screens[device->screenNo]->width;
    height = screenInfo.screens[device->screenNo]->height;

    DBG(10, xf86Msg(X_INFO, "%s: width %d,height %d,xSize %d,ySize %d \n",
                    methodName, width,height,xSize,ySize));

    *x = (v0 * width)  / xSize;
    *y = (v1 * height) / ySize;

    /* Deal with coordinate inversion
*/
    if ((device->flags & INVX_FLAG) != 0)
    {
        *x = width - *x;
    }
    if ((device->flags & INVY_FLAG) != 0)
    {
        *y = height - *y;
    }

    /* Normalize the adjusted sizes.
*/
    if (*x < 0)
    {
        *x = 0;
    }
    if (*x > width)
    {
        *x = width;
    }

    if (*y < 0)
    {
        *y = 0;
    }
    if (*y > height)
    {
        *y = height;
    }

    xf86XInputSetScreen(local, device->screenNo, *x, *y);

    DBG(20, xf86Msg(X_INFO, "%s: ends successfully, with: x=%d, y=%d\n",
                    methodName, *x, *y));

    /* return Success; */
    return 1;
}

/***********************************************************************
* PrometheanReverseConvert --
*
* Convert X and Y to valuators.
*/
static Bool
        PrometheanReverseConvert(
                LocalDevicePtr local,
                int            x,
                int            y,
                int*           valuators)
{
    const char*     methodName = "PrometheanReverseConvert";
    PrometheanDevicePtr device = (PrometheanDevicePtr) local->private;
    int             xSize, ySize;

    CHECK_LOCAL_DEVICE_PTR(local, __LINE__);
    CHECK_PROMETHEAN_DEVICE_PTR(device, __LINE__);

    DBG(20, xf86Msg(X_INFO, "%s: begins, with: x=%d, y=%d, "
                    "valuators[0]=%d, valuators[1]=%d\n",
                    methodName, x, y, valuators[0], valuators[1]));

    /* Adjust by tablet ratio
*/
    xSize = device->xBottom - device->xTop;
    ySize = device->yBottom - device->yTop;

    valuators[0] = (x*xSize) / screenInfo.screens[device->screenNo]->width;
    valuators[1] = (y*ySize) / screenInfo.screens[device->screenNo]->height;

    DBG(20, xf86Msg(X_INFO, "%s: Converted x,y (%d, %d) to (%d, %d)\n",
                    methodName, x, y, valuators[0], valuators[1]));

    xf86XInputSetScreen(local,device->screenNo,valuators[0], valuators[1]);
    DBG(10, xf86Msg(X_INFO, "%s: ends successfully, with valuators(%d,%d)\n",
                    methodName, valuators[0], valuators[1]));

    return Success;
}

/***************************************************************************
* PrometheanReadInput --
*
* Read the new events from the device, and enqueue them. Since a singular
* file descriptor is associated with up to three XFree86 devices, any
* of those three drivers may be awaken and dispatched by X's select loop.
*
* 1) The Cursor device therefore might receive events for the Eraser;
*    it must be aware of how to dispatch to more devices than itself.
* 2) The Cursor needs to the aware that if there is no Eraser device
*    associated with this physical tablet, it needs to throw out Eraser
*    events.
*/
static void
        PrometheanReadInput(
                LocalDevicePtr local)
{
    const char*         methodName = "PrometheanReadInput";
    PrometheanDevicePtr     device = (PrometheanDevicePtr) local->private;
    PrometheanCommonPtr     common = device->common;
    struct input_event* event;
    ssize_t             currRead;
    ssize_t             totalRead;
    int                 i, id;
    char                eventBuf[sizeof(struct input_event) * MAX_EVENTS];
    int                 eventsInMessage;
    double              d_z;
    double              d_zCapacity;
    int                 deviceIdArray[] = {STYLUS_ID, CURSOR_ID, ERASER_ID,
                                           STYLUS_ID, CURSOR_ID, ERASER_ID};
    int                 start, j;

    CHECK_LOCAL_DEVICE_PTR(local, __LINE__);
    CHECK_PROMETHEAN_DEVICE_PTR(device, __LINE__);
    CHECK_PROMETHEAN_COMMON_PTR(common, __LINE__);

    /* Read up to MAX_EVENTS from the tablet. Disallow any partial reads
* (e.g., we don't want any fractional input_events). We'll continue
* to read until we get at least one full input_event. The likelihood
* of being caught in here not reading a full input_event is low, given
* that we're dispatched by select(), and we know that on the kernel
* side, only full input_events are emitted. Good paranoia.
*/

    /* plugged out; close the device and loop until it comes back or until we want to exit */
    /*
if (PrometheanIsValidFileDescriptor(local->fd) == BadMatch)
{
xf86Msg(X_INFO, "%s: retry\n",methodName);
PrometheanOpenFileDescriptor(local);
if (local->fd == VALUE_NA)
{
sleep(1);
return;
}
}
*/

    totalRead = 0;
    do
    {
        currRead = -1;
        if (xf86WaitForInput(local->fd, 0))
        {
            SYSCALL(currRead,
                    xf86ReadSerial(local->fd, eventBuf + totalRead, sizeof(eventBuf) - totalRead));
        }
        if (currRead < 0)
        {
            int err = errno;
            xf86Msg(X_INFO, "%s: abends: read: %d, %s\n",
                    methodName,
                    err,
                    strerror(err));

            return;
        }

        totalRead += currRead;
    } while ((totalRead % sizeof(struct input_event)) != 0);

    /* Parse the events read from the tablet.
*
* We switch from one device to another through the auspices
* of button events. We'll allow that from the standpoint that
* we'll allow device-selecting events only if the underlying
* driver for that device is loaded/opened.
*
* All positional reports are handled as absolute reports.
* If they are relative, we adjust them into absolute.
*
* Later, we'll convert those reports back to relative if XFree
* has the driver configured in relative mode. This allows for the
* kernel driver and the XFree driver to be out of sync WRT
* absolute/relative coordinates, and yet still work.
*
* Actually, it allows the XFree user to set his tablet to relative
* mode without the worry of also going to the kernel driver and
* configuring it to agree with the XFree driver.
*/
    eventsInMessage = 0;
    for (event=(struct input_event *)(eventBuf);
    event<(struct input_event *)(eventBuf+totalRead);
    ++event)
    {
        /* Unprocessed events:
* ABS_RZ       - Rotate Stylus
* ABS_DISTANCE - Related to the Wheel?
* ABS_THROTTLE - Related to the Wheel?
*
* Synthesized events
* ABS_X_TILT - The promethean tablet does not report these,
* ABS_Y_TILT - but the Linux kernel driver sends synthetic values.
* ABS_WHEEL  - Ditto.
* REL_WHEEL  - Ditto.
*/
        /*        do
{ */
        switch (event->type)
        {
        case EV_ABS:
            {
                switch (event->code)
                {
                case ABS_X:
                    {
                        ++eventsInMessage;
                        common->currentValues.x = event->value;
                    }
                    break;

                case ABS_Y:
                    {
                        ++eventsInMessage;
                        common->currentValues.y = event->value;
                    }
                    break;

                case ABS_PRESSURE:
                    {
                        ++eventsInMessage;
                        common->currentValues.z = event->value;
                    }
                    break;

                case ABS_TILT_X:
                case ABS_RZ:
                    {
                        ++eventsInMessage;
                        common->currentValues.xTilt = event->value;
                    }
                    break;

                case ABS_TILT_Y:
                    {
                        ++eventsInMessage;
                        common->currentValues.yTilt = event->value;
                    }
                    break;

                case ABS_DISTANCE:
                    {
                        ++eventsInMessage;
                        common->currentValues.distance = event->value;
                    }
                    break;

                case ABS_WHEEL:
                case ABS_THROTTLE:
                    {
                        ++eventsInMessage;
                        common->currentValues.wheel = event->value;
                    }
                    break;

                case ABS_MISC:
                    {
                        /* We have an agreement with the
* Linux Promethean HID driver to send
* the proximity bit as bit 0 of the ABS_MISC value.
*
* The tool through which we get the report,
* the tablet's Stylus or Mouse device, is mapped
* to bits 4..7. Presently, the X driver does not
* use that information. Which means, if the driver
* is presently routing all events to, say,
* CURSOR_ID, then we don't mind the user actually
* using the stylus tool on the tablet for it.
*
* But we reserve the right to change our minds :->
*/
                        ++eventsInMessage;
                        common->currentValues.proximity =
                                PROXIMITY(event->value) != 0 ? 1 : 0;
                    }
                    break;
                }
            }
            break; /* EV_ABS */

        case EV_REL:
            {
                switch (event->code)
                {
                case REL_X:
                    {
                        /* Normalize all relative events into absolute
* coordinates.
*/
                        ++eventsInMessage;
                        common->currentValues.x =
                                common->previousValues.x + event->value;
                    }
                    break;

                case REL_Y:
                    {
                        /* Normalize all relative events into absolute
* coordinates.
*/
                        ++eventsInMessage;
                        common->currentValues.y =
                                common->previousValues.y + event->value;
                    }
                    break;

                case REL_WHEEL:
                    {
                        /* Normalize all relative events into absolute
* coordinates.
*/
                        ++eventsInMessage;
                        common->currentValues.wheel =
                                common->previousValues.wheel + event->value;
                    }

                case REL_MISC:
                    {
                        /* We have an agreement with the
* Linux Promethean HID driver to send
* the proximity bit through REL_MISC.
* We do this solely if proximity is
* being reported through the Mouse tool;
* else, if stylus, we'll get proximity through
* bits 0..3 of the ABS_MISC report.
*
* The tool through which we get the report,
* the tablet's Stylus or Mouse device, is mapped
* to bits 4..7. Presently, the X driver does not
* use that information. Which means, if the driver
* is presently routing all events to, say,
* CURSOR_ID, then we don't mind the user actually
* using the stylus tool on the tablet for it.
*
* But we reserve the right to change our minds :->
*/
                        ++eventsInMessage;
                        common->currentValues.proximity =
                                PROXIMITY(event->value) != 0 ? 1 : 0;
                    }
                    break;
                }
            }
            break; /* EV_REL */

        case EV_KEY:
            {
                switch (event->code)
                {
                    /* Events that begin with btn_tool_pen ..
* btn_tool_airbrush are from the stylus. Further, they
* indicate that the user wants input to be dispatched
* to the STYLUS_ID device. Which, we'll do, IF said device
* is presently open. Otherwise, no, we won't, we'll stay
* with the current device.
*/
                case BTN_TOOL_PEN:
                case BTN_TOOL_PENCIL:
                case BTN_TOOL_BRUSH:
                case BTN_TOOL_AIRBRUSH:
                    {
                        if (PrometheanIsEventValid(common, STYLUS_ID) == Success)
                        {
                            ++eventsInMessage;
                            common->currentValues.eventType = STYLUS_ID;
                        }
                        ++eventsInMessage;
                        common->currentValues.proximity = event->value;
                    }
                    break;

                    /* Events that begin with btn_tool_rubber are from the
* eraser. Further, they indicate that the user wants
* input to be dispatched to the ERASER_ID device. Which,
* we'll do, IF said device is presently open. Otherwise,
* no, we won't, we'll stay with the current device.
*/
                case BTN_TOOL_RUBBER:
                    {
                        if (PrometheanIsEventValid(common, ERASER_ID) == Success)
                        {
                            ++eventsInMessage;
                            common->currentValues.eventType = ERASER_ID;
                        }
                    }
                    break;

                    /* Events that begin with btn_tool_mouse-btn_tool_lens
* are from the cursor. Further, they indicate that the
* user wants input to be dispatched to the CURSOR_ID
* device. Which, we'll do, IF said device is presently
* open. Otherwise, no, we won't, we'll stay with the
* current device.
*/
                case BTN_TOOL_MOUSE:
                case BTN_TOOL_LENS:
                    {
                        if (PrometheanIsEventValid(common, CURSOR_ID) == Success)
                        {
                            ++eventsInMessage;
                            common->currentValues.eventType = CURSOR_ID;
                        }
                    }
                    break;

                    /* Remaining button events are just that -- button events.
* We get them, we report them.
*/
                case BTN_TOUCH:
                    {
                        ++eventsInMessage;
                        if (event->value > 0)
                        {
                            common->currentValues.button |=
                                    BUTTONS_EVENT_TOUCH;
                        }
                        else
                        {
                            common->currentValues.button &=
                                    ~BUTTONS_EVENT_TOUCH;
                        }
                    }
                    break;

                case BTN_STYLUS:
                    {
                        ++eventsInMessage;
                        if (event->value != 0)
                        {
                            common->currentValues.button |=
                                    BUTTONS_EVENT_STYLUS;
                        }
                        else
                        {
                            common->currentValues.button &=
                                    ~BUTTONS_EVENT_STYLUS;
                        }
                    }
                    break;

                case BTN_STYLUS2:
                    {
                        ++eventsInMessage;
                        if (event->value != 0)
                        {
                            common->currentValues.button |=
                                    BUTTONS_EVENT_STYLUS2;
                        }
                        else
                        {
                            common->currentValues.button &=
                                    ~BUTTONS_EVENT_STYLUS2;
                        }
                    }
                    break;

                    /* Normal Mouse button handling: LEFT, RIGHT and
* MIDDLE all buttons that we'll report to X
* as normal buttons. Note that the damned things
* re-use the same bitmasks as the Stylus buttons,
* above.
*/
                case BTN_LEFT:
                    {
                        ++eventsInMessage;
                        if (event->value != 0)
                        {
                            common->currentValues.button |=
                                    BUTTONS_EVENT_MOUSE_LEFT;
                        }
                        else
                        {
                            common->currentValues.button &=
                                    ~BUTTONS_EVENT_MOUSE_LEFT;
                        }
                    }
                    break;

                case BTN_MIDDLE:
                    {
                        ++eventsInMessage;
                        if (event->value != 0)
                        {
                            common->currentValues.button |=
                                    BUTTONS_EVENT_MOUSE_MIDDLE;
                        }
                        else
                        {
                            common->currentValues.button &=
                                    ~BUTTONS_EVENT_MOUSE_MIDDLE;
                        }
                    }
                    break;

                case BTN_RIGHT:
                    {
                        ++eventsInMessage;
                        if (event->value != 0)
                        {
                            common->currentValues.button |= BUTTONS_EVENT_MOUSE_RIGHT;
                        }
                        else
                        {
                            common->currentValues.button &= ~BUTTONS_EVENT_MOUSE_RIGHT;
                        }
                    }
                    break;

                case BTN_SIDE:
                    {
                        ++eventsInMessage;
                        if (event->value != 0)
                        {
                            common->currentValues.button |=
                                    BUTTONS_EVENT_SIDE_BTN;
                        }
                        else
                        {
                            common->currentValues.button &=
                                    ~BUTTONS_EVENT_SIDE_BTN;
                        }
                    }
                    break;

                case BTN_EXTRA:
                    {
                        ++eventsInMessage;
                        if (event->value != 0)
                        {
                            common->currentValues.button |=
                                    BUTTONS_EVENT_EXTRA_BTN;
                        }
                        else
                        {
                            common->currentValues.button &=
                                    ~BUTTONS_EVENT_EXTRA_BTN;
                        }
                    }
                    break;

                    /* Any other EV_KEY event is assumed to be
* the pressing of a macro key from the tablet.
*/
                default:
                    {
                        ++eventsInMessage;
                        common->currentValues.macroKey = event->value;
                    }
                    break;
                }
            }
            break; /* EV_KEY */

        } /* switch event->type */

        /* We have two potential event terminators. EV_MSC was used
* by (unwritten) convention to indicate the end-of-report.
* Problem is, EV_MSC is supposed to actually report data,
* so a new event type, EV_SYN, was created in Linux 2.5.x
* expressively for this purpose.
*
* Theoretically, if EV_SYN is defined, we should only terminate
* the population of device->currentValues struct IFF we receive
* that event. The fact of the matter is, the EV_MSC is assumed
* to be an ugliness that will take some time to be deprecated.
* For the nonce, we'll support both. But, if you have a tablet
* that's actually supplying something interesting with EV_MSC,
* this is obviously some code that will require modifications.
*/
        /*  } while (event->type != EV_MSC && event->type != EV_SYN); */
        if ((event->type == EV_MSC || event->type == EV_SYN) &&
            eventsInMessage) {

            /* We've seen EV_MSCs in the incoming data trail with no
* other message types in-between. We use 'eventsInMessage'
* to count all 'real' messages in-between. If there were none,
* do NOT copy common->currentValues to common->previousValues
* (as this will make the jitter filter useless). Just go and
* read the subsequent events.
*/
            /*        if (eventsInMessage == 0)
{
continue;
}
eventsInMessage = 0; */

            /* This filter throws out reports that do not meet minimum threshold
* requirements for movement along that axis.
*
* Presently, we discard the entire report if any dimension of the
* currentValues struct does not meet it's minimum threshold.
*
* Also, we only do the comparison IFF a threshold has been set
* for that given dimension.
*/
            if ((device->xThreshold > 1 &&
                 ABS(common->currentValues.x - common->previousValues.x)
                 <= device->xThreshold) ||
                (device->yThreshold > 1 &&
                 ABS(common->currentValues.y - common->previousValues.y)
                 <= device->yThreshold) ||
                (device->zThreshold > 1 &&
                 ABS(common->currentValues.z - common->previousValues.z)
                 <= device->zThreshold))
            {
                DBG(10, xf86Msg(X_INFO, "%s: Event Filtered Out by Thresholds\n",
                                methodName));
                continue;
            }

            /* If this report somehow has exactly the same readings as the
* previous report for all dimensions, throw the report out.
*/
            if ((common->currentValues.x == common->previousValues.x) &&
                (common->currentValues.y == common->previousValues.y) &&
                (common->currentValues.z == common->previousValues.z) &&
                (common->currentValues.proximity ==
                 common->previousValues.proximity) &&
                (common->currentValues.button ==
                 common->previousValues.button) &&
                (common->currentValues.macroKey ==
                 common->previousValues.macroKey))
            {
                DBG(10, xf86Msg(X_INFO, "%s: Event Filtered Out\n",
                                methodName));
                continue;
            }

            /* We have three different methods by which we report pressure, Z.
* One is to use linear values from 0 to common->zCapacity. The
* other two, SOFT_SMOOTH and HARD_SMOOTH, use different
* algorithms to 'smooth out' the values.
*/
            d_z         = (double)common->currentValues.z;
            d_zCapacity = (double)common->zMaxCapacity + 1;

            switch (device->zMode)
            {
            case VALUE_NA:
            case PRESSURE_MODE_LINEAR:
                {
                    /* Leave Z alone. */
                }
                break;

            case PRESSURE_MODE_SOFT_SMOOTH:
                {
                    /*	    d_z = (d_z * d_z / d_zCapacity)+ 0.5; */
                }
                break;

            case PRESSURE_MODE_HARD_SMOOTH:
                {
                    /*	    d_z = (d_zCapacity * sqrt( d_z / d_zCapacity)) + 0.5; */
                }
                break;
            }
            common->currentValues.smooth_z = (int)d_z;

            /* What do we do if we have an event for a device that
* is closed? Do we throw it out or make an attempt to reroute
* to another device? We will attempt to reroute it. Otherwise, the
* tablet becomes silently inert, which isn't good.
*
* As we're using powers of 2 for values in deviceIdArray,
* figuring out the starting element thus works fine.
*/
            start = common->currentValues.eventType / 2;

            /* This allows us to track that all devices are closed.
*/
            common->currentValues.eventType = 0;

            for (i = deviceIdArray[start], j = 0;
                 j < PROMETHEAN_MAX_DEVICES;
                 i = deviceIdArray[j], ++j)
            {
                if (PrometheanIsEventValid(common, i) == Success)
                {
                    common->currentValues.eventType = /* 1 << */ i;
                    break;
                }
            }
            /* What if ALL devices are closed? Well, pretend it's a stylus,
* then...
*/
            if (common->currentValues.eventType == 0)
            {
                common->currentValues.eventType = STYLUS_ID;
            }
            /* Put up a message if we re-route a message.
*/
            if (deviceIdArray[start] != deviceIdArray[j])
            {
                DBG(5, xf86Msg(X_INFO, "%s: Event for closed device! Rerouted\n",
                               methodName));
            }

            /* Dispatch events to one of our configured devices.
*/
            for (i=0; i < PROMETHEAN_MAX_DEVICES; ++i)
            {
                /* Do not consider processing null device slots or
* device drivers that are not open.
*/
                if (common->deviceArray[i] != NULL &&
                    PrometheanIsValidFileDescriptor(common->deviceArray[i]->fd) == Success)
                {
                    id = ((PrometheanDevicePtr)common->deviceArray[i]->private)->deviceId;
                    if (DEVICE_ID(common->currentValues.eventType) == DEVICE_ID(id))
                    {
                        PrometheanSendEvents(common->deviceArray[i]);
                        break;
                    }
                }
            }

            memmove(&common->previousValues, &common->currentValues,
                    sizeof(PrometheanStateRec));
            common->currentValues.macroKey = VALUE_NA;
        }
    }
}

/***********************************************************************
* PrometheanIsEventValid --
*
* Return indication that the event for device 'x' is valid. Validity is
* defined as an XFree86 driver instance for the given device, and that said
* driver is in an open state.
*/
static Bool
        PrometheanIsEventValid(PrometheanCommonPtr common, int code)
{
    int i, id;

    CHECK_PROMETHEAN_COMMON_PTR(common, __LINE__);

    for (i=0; i < PROMETHEAN_MAX_DEVICES; ++i)
    {
        if (common->deviceArray[i] != NULL &&
            PrometheanIsValidFileDescriptor(common->deviceArray[i]->fd) == Success)
        {
            id = ((PrometheanDevicePtr)common->deviceArray[i]->private)->deviceId;
            if (DEVICE_ID(code) == DEVICE_ID(id))
            {
                return Success;
            }
        }
    }
    return BadMatch;
}

/**********************************************************************
* PrometheanSendEvents --
*
* Send events according to the device state.
*/
static void
        PrometheanSendEvents(
                LocalDevicePtr local)
{
    const char*     methodName = "PrometheanSendEvents";
    PrometheanDevicePtr device = (PrometheanDevicePtr) local->private;
    PrometheanCommonPtr common = device->common;
    int             bCorePointer, bAbsolute;
    int             x, y, z;
    int             delta, id;
    const int numAxes = 3;

    CHECK_LOCAL_DEVICE_PTR(local, __LINE__);
    CHECK_PROMETHEAN_DEVICE_PTR(device, __LINE__);
    CHECK_PROMETHEAN_COMMON_PTR(common, __LINE__);
    DBG(20,xf86Msg(X_INFO, "%s: starting the method\n", methodName));

    if (DEVICE_ID(device->deviceId) != DEVICE_ID(common->currentValues.eventType))
    {
        xf86Msg(X_INFO, "%s: not the same device type (%u,%u)\n",
                methodName,
                DEVICE_ID(device->deviceId), common->currentValues.eventType);
        return;
    }

    bAbsolute    = (device->flags & ABSOLUTE_FLAG) ? 1 : 0;
    bCorePointer = 0;
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
    bCorePointer = xf86IsCorePointer(local->dev);
#endif
    DBG(20,xf86Msg(X_INFO, "%s: isCorePointer ? %d\n", methodName,bCorePointer));
    DBG(20,xf86Msg(X_INFO, "%s: isAbsolute ? %d\n", methodName,bAbsolute));

    /* Normalize X and Y coordinates. This includes dealing
* with absolute/relative coordinate mode.
*/
    if (bAbsolute != 0)
    {
        x = common->currentValues.x;
        y = common->currentValues.y;
    }
    else
    {
        x = common->currentValues.x - common->previousValues.x;
        y = common->currentValues.y - common->previousValues.y;
    }
    z     = common->currentValues.smooth_z;

    /* Translate coordinates according to Top and Bottom points.
*/
    if (x > device->xBottom)
    {
        x = device->xBottom;
    }

    if (y > device->yBottom)
    {
        y = device->yBottom;
    }

    if (device->xTop > 0)
    {
        DBG(5, xf86Msg(X_INFO, "%s: Adjusting x, with xTop=%d\n",
                       methodName, device->xTop));
        x = device->xTop;
    }

    if (device->yTop > 0)
    {
        DBG(5, xf86Msg(X_INFO, "%s: Adjusting y, with yTop=%d\n",
                       methodName, device->yTop));
        y = device->yTop;
    }

    if (x < 0)
    {
        x = 0;
    }
    if (y < 0)
    {
        y = 0;
    }

    /* Deal with pressure min..max, which differs from threshold.
*/
    if (z < device->zMin)
    {
        z = 0;
    }

    if (z > device->zMax)
    {
        z = device->zMax;
    }

    /* As the coordinates are ready, we can send events to X */
    DBG(20,xf86Msg(X_INFO, "%s: CurrentValues.proximity: %d\n", methodName,common->currentValues.proximity));

    if (common->currentValues.proximity != 0)
    {
        DBG(20,xf86Msg(X_INFO, "%s: current values are proximity \n", methodName));
        if (common->previousValues.proximity == 0)
        {
            DBG(20,xf86Msg(X_INFO, "%s: previous values were not proximity \n", methodName));
            if (bCorePointer == 0)
            {

                DBG(15,xf86Msg(X_INFO, "%s: Posting proximity event \n", methodName));
                xf86PostProximityEvent(local->dev, 1, 0, numAxes, x, y, z);
            }
        }

        if ((bAbsolute != 0 &&
             (common->previousValues.x != common->currentValues.x   ||
              common->previousValues.y != common->currentValues.y   ||
              common->previousValues.z != common->currentValues.z)) ||
            bAbsolute == 0)
        {
            if (bAbsolute == 0)
            {
                ++x;
                ++y;
            }

            DBG(15,xf86Msg(X_INFO, "%s: Posting Motion event x %d y %d \n", methodName,x,y));

            xf86PostMotionEvent(local->dev, bAbsolute != 0, 0, numAxes, x, y, z);
        }

        if (common->previousValues.button != common->currentValues.button)
        {
            delta = common->currentValues.button ^
                    common->previousValues.button;
            while (delta != 0)
            {
                id = ffs(delta);
                delta &= ~(1 << (id-1));
                DBG(15,xf86Msg(X_INFO, "%s: Posting Button event id = %d, press %d \n", methodName,id,
                               (common->currentValues.button & (1<<(id-1)))?1:0));

                xf86PostButtonEvent(local->dev, bAbsolute != 0, id,
                                    (common->currentValues.button & (1<<(id-1)))?1:0, 0,numAxes, x, y, z);
            }
        }
    }
    else
    {
        if (bCorePointer == 0)
        {
            if (common->previousValues.proximity != 0)
            {
                DBG(15,xf86Msg(X_INFO, "%s: Posting Out Of Proximity\n", methodName));
                xf86PostProximityEvent(local->dev, 0, 0, numAxes, x, y, z);
            }
        }
        common->previousValues.proximity = 0;
    }
}


/***********************************************************************
* Plug, unplug, and structs are put at the end of the driver. Why is
* unknown at present time.
*/
static
        InputDriverRec PROMETHEAN =
{
    1,                  /* driver version */
    "promethean",           /* driver name */
    NULL,               /* identify */
    PrometheanInit,         /* pre-init */
    PrometheanUninit,       /* un-init */
    NULL,               /* module */
    0                   /* ref count */
};

/***********************************************************************
* PrometheanUnplug --
*
* Should remove all vestiges of the input driver. TODO.
*/
static void
        PrometheanUnplug(
                pointer p)
{
    ErrorF("PrometheanUnplug\n");
    xf86Msg(X_INFO, "Promethean UNPlug: Promethean USB X Driver Version 5.2.4\n");
}

/***********************************************************************
* PrometheanPlug --
*
* Called to load our driver into memory, do registration, etc., processing
*/
static pointer
        PrometheanPlug(
                pointer    module,
                pointer    options,
                int*       errmaj,
                int*       errmin)
{
    xf86AddInputDriver(&PROMETHEAN, module, 0);
    xf86Msg(X_INFO, "PrometheanPlug: Promethean USB X Driver Version 5.2.4\n");
    return module;
}

/* Version ID and module identification "stuff"
*/
static
        XF86ModuleVersionInfo PrometheanVersionRec =
{
    "promethean",
    MODULEVENDORSTRING,
    MODINFOSTRING1,
    MODINFOSTRING2,
#ifdef PROMETHEAN_XORG_1_6
    XORG_VERSION_CURRENT,
#else
    XF86_VERSION_CURRENT,
#endif
    DRIVER_MAJOR, DRIVER_MINOR, 0,
    ABI_CLASS_XINPUT,
    ABI_XINPUT_VERSION,
    MOD_CLASS_XINPUT,
    {0, 0, 0, 0}        /* A signature that we ignore */
};

#ifdef _X_EXPORT
_X_EXPORT XF86ModuleData prometheanModuleData =
#else
        XF86ModuleData prometheanModuleData =
#endif
{
    &PrometheanVersionRec,
    PrometheanPlug,
    PrometheanUnplug
};

/*======================================================================
*
* End of xf86Promethean.c
*/

